# This file is part of Xpra.
# Copyright (C) 2014-2023 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
from queue import Queue, Empty
from typing import Tuple, Dict, Any, Callable, Optional

from xpra.util import typedict, envint, parse_simple_dict
from xpra.os_util import OSX
from xpra.gst_common import import_gst, GST_FLOW_OK
from xpra.gst_pipeline import Pipeline
from xpra.log import Logger

Gst = import_gst()
log = Logger("encoder", "gstreamer")

FRAME_QUEUE_TIMEOUT = envint("XPRA_GSTREAMER_FRAME_QUEUE_TIMEOUT", 1)
FRAME_QUEUE_INITIAL_TIMEOUT = envint("XPRA_GSTREAMER_FRAME_QUEUE_INITIAL_TIMEOUT", 3)


def get_default_encoder_options() -> Dict[str,Dict[str,Any]]:
    options : Dict[str,Dict[str,Any]] = {
        "vaapih264enc" : {
            "max-bframes"   : 0,    #int(options.boolget("b-frames", False))
            #"tune"          : 3,    #low-power
            #"rate-control" : 8, #qvbr
            "compliance-mode" : 0,  #restrict-buf-alloc (1) – Restrict the allocation size of coded-buffer
            #"keyframe-period"   : 9999,
            #"prediction-type" : 1, #hierarchical-p (1) – Hierarchical P frame encode
            #"quality-factor" : 10,
            #"quality-level" : 50,
            #"bitrate"   : 2000,
            #"prediction-type" : 1,    #Hierarchical P frame encode
            #"keyframe-period" : 4294967295,
            "aud"   : True,
            },
        "vaapih265enc" : {
            "max-bframes"   : 0,    #int(options.boolget("b-frames", False))
            #"tune"          : 3,    #low-power
            #"rate-control" : 8, #qvbr
            },
        "amfh264enc" : {
            "usage"    : "ultra-low-latency",
            },
        "amfh265enc" : {
            "usage"    : "ultra-low-latency",
            },
        "x264enc" : {
            "speed-preset"  : "ultrafast",
            "tune"          : "zerolatency",
            "byte-stream"   : True,
            "threads"       : 1,
            "key-int-max"   : 15,
            "intra-refresh" : True,
            },
        "vp8enc" : {
            "deadline"      : 1,
            "error-resilient" : 0,
            },
        "vp9enc" : {
            "deadline"      : 1,
            "error-resilient" : 0,
            "lag-in-frames" : 0,
            "cpu-used"      : 16,
            },
        "nvh264enc" : {
            "zerolatency"   : True,
            "rc-mode"       : 3,    #vbr
            "preset"        : 5,    #low latency, high performance
            "bframes"       : 0,
            "aud"           : True,
            },
        "nvh265enc" : {
            "zerolatency"   : True,
            "rc-mode"       : 3,    #vbr
            "preset"        : 5,    #low latency, high performance
            #should be in GStreamer 1.18, but somehow missing?
            #"bframes"       : 0,
            "aud"           : True,
            },
        "nvd3d11h264enc" : {
            "bframes"       : 0,
            "aud"           : True,
            "preset"        : 5,    #low latency, high performance
            "zero-reorder-delay"    : True,
            },
        "nvd3d11h265enc" : {
            "bframes"       : 0,
            "aud"           : True,
            "preset"        : 5,    #low latency, high performance
            "zero-reorder-delay"    : True,
            },
        "svtav1enc" : {
            # "speed"         : 12,
            # "gop-size"      : 251,
            "intra-refresh" : 1,    #open gop
            # "lookahead"     : 0,
            # "rc"            : 1,    #vbr
            },
        "svtvp9enc" : {
            },
        #"svthevcenc" : {
        #    "b-pyramid"         : 0,
        #    "baselayer-mode"    : 1,
        #    "enable-open-gop"   : True,
        #    "key-int-max"       : 255,
        #    "lookahead"         : 0,
        #    "pred-struct"       : 0,
        #    "rc"                : 1, #vbr
        #    "speed"             : 9,
        #    "tune"              : 0,
        #    }
        }
    if not OSX:
        options["av1enc"] = {
            "cpu-used"          : 5,
            "end-usage"         : 2,    #cq
            }
    #now apply environment overrides:
    for element in options.keys():
        enc_options_str = os.environ.get(f"XPRA_{element.upper()}_OPTIONS", "")
        if enc_options_str:
            encoder_options = parse_simple_dict(enc_options_str)
            log(f"user overridden options for {element}: {encoder_options}")
            options[element] = encoder_options
    return options

def get_default_decoder_options() -> Dict[str,Dict[str,Any]]:
    options : Dict[str,Dict[str,Any]] = {
        "av1dec"    : {
            "stream-format": "obu-stream",
            "alignment": "tu",
        },
    }
    return options


def get_version() -> Tuple[int, ...]:
    return (5, 0)

def get_type() -> str:
    return "gstreamer"

def get_info() -> Dict[str,Any]:
    return {"version"   : get_version()}

def init_module() -> None:
    log("gstreamer.init_module()")

def cleanup_module() -> None:
    log("gstreamer.cleanup_module()")


def get_gst_rgb_format(rgb_format : str) -> str:
    if rgb_format in (
        "NV12",
        "RGBA", "BGRA", "ARGB", "ABGR",
        "RGB", "BGR",
        "RGB15", "RGB16", "BGR15",
        "r210",
        "BGRP", "RGBP",
        ):
        #identical name:
        return rgb_format
    #translate to gstreamer name:
    return {
        "YUV420P"   : "I420",
        "YUV444P"   : "Y444",
        "BGRX"      : "BGRx",
        "XRGB"      : "xRGB",
        "XBGR"      : "xBGR",
        "YUV400"    : "GRAY8",
        #"RGB8P"
        }[rgb_format]


def get_video_encoder_caps(encoder:str="x264enc") -> Dict[str,Any]:
    if encoder=="jpeg":
        return {}
    if encoder=="av1enc":
        return {
            "alignment"     : "tu",
            "stream-format" : "obu-stream",
            }
    return {
        "alignment"     : "au",
        "stream-format" : "byte-stream",
        }

def get_video_encoder_options(encoder:str="x264", profile:str="", options:Optional[typedict]=None):
    eopts = get_default_encoder_options().get(encoder, {})
    eopts["name"] = "encoder"
    if encoder=="x264enc" and options:
        from xpra.codecs.codec_constants import get_x264_quality, get_x264_preset
        q = get_x264_quality(options.intget("quality", 50), profile)
        s = options.intget("speed", 50)
        eopts.update({
            "pass"  : "qual",
            "quantizer" : q,
            "speed-preset" : get_x264_preset(s),
            })
    #should check for "bframes" flag in options?
    return eopts

def get_gst_encoding(encoding:str) -> str:
    if encoding in ("jpeg", "png"):
        return f"image/{encoding}"
    video = {"hevc" : "h265"}.get(encoding, encoding)
    return f"video/x-{video}"


class VideoPipeline(Pipeline):
    __generic_signals__ : Dict[str,Tuple] = Pipeline.__generic_signals__.copy()
    """
    Dispatch video encoding or decoding to a gstreamer pipeline
    """
    def init_context(self, encoding:str, width:int, height:int, colorspace:str, options=None):
        options = typedict(options or {})
        self.encoding : str = encoding
        self.width : int = width
        self.height : int = height
        self.colorspace : str = colorspace
        self.frames : int = 0
        self.frame_queue : Queue[Any] = Queue()
        self.pipeline_str : str = ""
        self.create_pipeline(options)
        self.src : Gst.Element = self.pipeline.get_by_name("src")
        self.src.set_property("format", Gst.Format.TIME)
        #self.src.set_caps(Gst.Caps.from_string(CAPS))
        self.sink : Gst.Element = self.pipeline.get_by_name("sink")
        def sh(sig:str, handler:Callable):
            self.element_connect(self.sink, sig, handler)
        sh("new-sample", self.on_new_sample)
        sh("new-preroll", self.on_new_preroll)
        self.start()

    def create_pipeline(self, options):
        raise NotImplementedError()

    def on_message(self, bus, message) -> int:
        if message.type == Gst.MessageType.NEED_CONTEXT and self.pipeline_str.find("vaapi")>=0:
            log("vaapi is requesting a context")
            return GST_FLOW_OK
        return super().on_message(bus, message)

    def on_new_preroll(self, _appsink) -> int:
        log("new-preroll")
        return GST_FLOW_OK

    def process_buffer(self, buf):
        r = self.src.emit("push-buffer", buf)
        if r!=GST_FLOW_OK:
            log.error("Error: unable to push image buffer")
            return None
        timeout = FRAME_QUEUE_INITIAL_TIMEOUT if self.frames==0 else FRAME_QUEUE_TIMEOUT
        try:
            return self.frame_queue.get(timeout=timeout)
        except Empty:
            log.error(f"Error: frame queue timeout after {timeout}s")
            try:
                btype = type(buf).__qualname__
                log.error(f" on {btype!r} of size {buf.get_size()}")
            except AttributeError:
                pass
            for k,v in self.get_info().items():
                log.error(f" {k:<16}: {v}")
            return None


    def get_info(self) -> Dict[str,Any]:
        info : Dict[str,Any] = get_info()
        if not self.colorspace:
            return info
        info.update({
            "frames"    : self.frames,
            "width"     : self.width,
            "height"    : self.height,
            "encoding"  : self.encoding,
            "colorspace": self.colorspace,
            "version"   : get_version(),
            })
        return info

    def __repr__(self):
        if not self.colorspace:
            return "gstreamer(uninitialized)"
        return f"gstreamer({self.colorspace} - {self.width}x{self.height})"

    def is_ready(self) -> bool:
        return bool(self.colorspace)

    def is_closed(self) -> bool:
        return not bool(self.colorspace)


    def get_encoding(self) -> str:
        return self.encoding

    def get_width(self) -> int:
        return self.width

    def get_height(self) -> int:
        return self.height

    def get_type(self) -> str:
        return "gstreamer"

    def clean(self) -> None:
        super().cleanup()
        self.width = 0
        self.height = 0
        self.colorspace = ""
        self.encoding = ""
        self.frames = 0


    def do_emit_info(self) -> None:
        self.emit_info_timer = 0



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2018-2023 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys
from io import BytesIO
from math import cos, sin
from typing import Type, Tuple, Dict, Any

from xpra.common import noop
from xpra.util import typedict, envint, AtomicInteger
from xpra.os_util import WIN32, load_binary_file, is_X11
from xpra.log import Logger
from xpra.platform.paths import get_icon_filename
from xpra.client.gui.fake_client import FakeClient

log = Logger("opengl", "paint")


def get_gl_client_window_module(force_enable=False) -> Tuple[Dict,Any]:
    from importlib import import_module
    try:
        opengl = import_module("OpenGL")
        log(f"opengl=%s", opengl)
    except ImportError as e:
        log("cannot import the OpenGL module", exc_info=True)
        log.warn("Warning: cannot import the 'OpenGL' module")
        log.warn(" %s", e)
        return {
            "success": False,
            "message": str(e),
        }, None
    log("get_gl_client_window_module()")
    try:
        from xpra.client.gl.gtk3 import nativegl_client_window
    except ImportError as e:
        log("cannot import opengl window module", exc_info=True)
        log.warn("Warning: cannot import native OpenGL module")
        log.warn(" %s", e)
        return {
            "success": False,
            "message": str(e),
        }, None
    opengl_props = nativegl_client_window.check_support(force_enable)
    log("check_support(%s)=%s", force_enable, opengl_props)
    if opengl_props:
        return opengl_props, nativegl_client_window
    return {
        "success": False,
        "message": "no valid OpenGL backend found",
    }, None


def get_test_gl_icon():
    data = b""
    encoding = "png"
    w = 32
    h = 32
    stride = w * 4
    gl_icon = get_icon_filename("opengl", ext="png")
    if gl_icon:
        try:
            from PIL import Image  # @UnresolvedImport pylint: disable=import-outside-toplevel
        except ImportError as e:
            log(f"testing without icon: {e}")
        else:
            img = Image.open(gl_icon)
            img.load()
            w, h = img.size
            stride = w * 4
            noalpha = Image.new("RGB", img.size, (255, 255, 255))
            noalpha.paste(img, mask=img.split()[3])  # 3 is the alpha channel
            buf = BytesIO()
            try:
                noalpha.save(buf, format="JPEG")
                data = buf.getvalue()
                buf.close()
                encoding = "jpeg"
            except KeyError as e:
                log("save()", exc_info=True)
                log.warn("OpenGL using png as jpeg is not supported by Pillow: %s", e)
                data = load_binary_file(gl_icon)
    if not data:
        data = bytes([0]) * stride * h
        encoding  = "rgb32"
    return encoding, w, h, stride, data

def no_idle_add(fn, *args, **kwargs):
    fn(*args, **kwargs)

def test_gl_client_window(gl_client_window_class : Type, max_window_size=(1024, 1024), pixel_depth=24, show=False):
    #try to render using a temporary window:
    draw_result = {}
    window = None
    try:
        x, y = -100, -100
        if show:
            x, y = 100, 100
        ww, wh = 250, 250
        from xpra.codecs.loader import load_codec
        load_codec("dec_pillow")
        from xpra.client.gui.window_border import WindowBorder
        border = WindowBorder()
        default_cursor_data = None
        noclient = FakeClient()
        #test with alpha, but not on win32
        #because we can't do alpha on win32 with opengl
        metadata = typedict({"has-alpha" : not WIN32})
        class NoHeaderGLClientWindow(gl_client_window_class):
            def add_header_bar(self):
                """ pretend to add the header bar """
            def schedule_recheck_focus(self):
                """ pretend to handle focus checks """
        window = NoHeaderGLClientWindow(noclient, None, 0, 2**32-1, x, y, ww, wh, ww, wh,
                                        metadata, False, typedict({}),
                                        border, max_window_size, default_cursor_data, pixel_depth)
        window_backing = window._backing
        window_backing.idle_add = no_idle_add
        window_backing.timeout_add = noop
        window_backing.source_remove = noop
        window.realize()
        window_backing.paint_screen = True
        pixel_format = "BGRX"
        options = typedict({"pixel_format" : pixel_format})
        widget = window_backing._backing
        widget.realize()
        def paint_callback(success, message=""):
            log("paint_callback(%s, %s)", success, message)
            draw_result["success"] = success
            if message:
                draw_result["message"] = message.replace("\n", " ")
        pix = AtomicInteger(0x7f)
        REPAINT_DELAY = envint("XPRA_REPAINT_DELAY", int(show)*16)

        coding, w, h, stride, icon_data = get_test_gl_icon()
        log("OpenGL: testing draw on %s widget %s with %s : %s", window, widget, coding, pixel_format)
        def draw():
            v = pix.increase()
            img_data = bytes([v % 256]*w*4*h)
            options["flush"] = 1
            window.draw_region(0, 0, w, h, "rgb32", img_data, w*4, v, options, [paint_callback])
            options["flush"] = 0
            mx = ww//2-w//2
            my = wh//2-h//2
            draw_x = round(mx*(1+sin(v/100)))
            draw_y = round(my*(1+cos(v/100)))
            log("calling draw_region for test gl icon")
            window.draw_region(draw_x, draw_y, w, h, coding, icon_data, stride, v, options, [paint_callback])
            return REPAINT_DELAY>0
        #the paint code is actually synchronous here,
        #so we can check the present_fbo() result:
        if show:
            widget.show()
            window.show()
            import gi
            gi.require_version("Gtk", "3.0")
            from gi.repository import Gtk, GLib  # @UnresolvedImport
            def window_close_event(*_args):
                Gtk.main_quit()
            noclient.window_close_event = window_close_event
            GLib.timeout_add(REPAINT_DELAY, draw)
            Gtk.main()
        else:
            draw()
            #ugly workaround for calling the paint handler
            #when the main loop is not running:
            if hasattr(window_backing, "on_realize"):
                window_backing.on_realize()
        if window_backing.last_present_fbo_error:
            return {
                "success" : False,
                "message" : "failed to present FBO on screen: %s" % window_backing.last_present_fbo_error
                }
    finally:
        if window:
            window.destroy()
    log("test_gl_client_window(..) draw_result=%s", draw_result)
    return draw_result or {"success" : False, "message" : "not painted on screen"}



def main(argv):
    from xpra.platform import program_context
    with program_context("opengl", "OpenGL"):
        try:
            if "-v" in argv or "--verbose" in argv:
                log.enable_debug()
            if is_X11():
                from xpra.x11.gtk3.gdk_display_source import init_gdk_display_source
                init_gdk_display_source()
            opengl_props, gl_client_window_module = get_gl_client_window_module(True)
            log("do_run_glcheck() opengl_props=%s, gl_client_window_module=%s", opengl_props, gl_client_window_module)
            gl_client_window_class = gl_client_window_module.GLClientWindow
            pixel_depth = 0
            log("do_run_glcheck() gl_client_window_class=%s, pixel_depth=%s", gl_client_window_class, pixel_depth)
            #if pixel_depth not in (0, 16, 24, 30) and pixel_depth<32:
            #    pixel_depth = 0
            draw_result = test_gl_client_window(gl_client_window_class, pixel_depth=pixel_depth, show=True)
            success = draw_result.pop("success", False)
            opengl_props.update(draw_result)
            if not success:
                opengl_props["safe"] = False
            return 0
        except Exception:
            log("do_run_glcheck(..)", exc_info=True)
            return 1

if __name__ == "__main__":
    r = main(sys.argv)
    sys.exit(r)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2012 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
# Copyright (C) 2012-2023 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Dict, Tuple

#These chipsets will use OpenGL,
#there will not be any warnings, even if the vendor is greylisted:
try:
    from typing_extensions import TypeAlias
    GL_MATCH_LIST : TypeAlias = Dict[str, Tuple[str,...]]
except ImportError:
    GL_MATCH_LIST = Dict

WHITELIST : GL_MATCH_LIST = {
    "renderer"  : (
        "Haswell", "Skylake", "Kabylake", "Cannonlake",
        "Whiskeylake", "Amberlake", "Cascadelake", "Cometlake",
        "Icelake", "Cooperlake",
    ),
}

#Chipsets from these vendors will trigger warnings,
#but OpenGL will still be enabled:
GREYLIST : GL_MATCH_LIST = {
    #"vendor"    : ("Intel", )
}

#Versions older than this will trigger warnings:
VERSION_REQ = {
   "nouveau" : (3, 0),      #older versions have issues
}

#These chipsets will be disabled by default:
BLACKLIST : GL_MATCH_LIST = {
    "renderer" :
        (
            "SVGA3D",
            "Software Rasterizer",
        ),
    "vendor"    : (
        #"VMware, Inc.",
        #"Humper",
        #to disable nvidia, uncomment this:
        #"NVIDIA Corporation",
    ),
    "platform"  : (
        #"darwin",
    ),
}


#for testing:
#GREYLIST["vendor"].append("NVIDIA Corporation")
#WHITELIST["renderer"] = ["GeForce GTX 760/PCIe/SSE2"]
#frequent crashes on OSX with GT 650M: (see ticket #808)
#if OSX:
#    GREYLIST.setdefault("vendor", []).append("NVIDIA Corporation")


class OpenGLFatalError(ImportError):
    pass



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2012 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
# Copyright (C) 2012-2023 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys
import logging
from typing import Any, Tuple, Dict, List

from xpra.util import envbool, envint, csv
from xpra.os_util import bytestostr
from xpra.log import Logger, CaptureHandler
from xpra.client.gl.gl_drivers import (
    GL_MATCH_LIST, WHITELIST, GREYLIST, VERSION_REQ, BLACKLIST,
    OpenGLFatalError,
)

log = Logger("opengl")

required_extensions : Tuple[str, ...] = ("GL_ARB_texture_rectangle", "GL_ARB_vertex_program")


GL_ALPHA_SUPPORTED : bool = envbool("XPRA_ALPHA", True)
DOUBLE_BUFFERED : bool = envbool("XPRA_OPENGL_DOUBLE_BUFFERED", True)

CRASH : bool = envbool("XPRA_OPENGL_FORCE_CRASH", False)
TIMEOUT : int = envint("XPRA_OPENGL_FORCE_TIMEOUT", 0)


#by default, we raise an ImportError as soon as we find something missing:
def raise_error(msg) -> None:
    raise ImportError(msg)
def raise_fatal_error(msg) -> None:
    raise OpenGLFatalError(msg)
gl_check_error = raise_error
gl_fatal_error = raise_fatal_error


def parse_pyopengl_version(vstr : str) -> Tuple[int,...]:
    def numv(s):
        try:
            return int(s)
        except ValueError:
            return 0
    return tuple(numv(x) for x in vstr.split("."))

_version_warning_shown = False


def check_functions(force_enable, *functions) -> None:
    missing = []
    available = []
    for x in functions:
        try:
            name = x.__name__
        except AttributeError:
            name = str(x)
        if not bool(x):
            missing.append(name)
        else:
            available.append(name)
    if missing:
        if not force_enable:
            raise_fatal_error("some required OpenGL functions are missing:\n%s" % csv(missing))
        log("some functions are missing: %s", csv(missing))
    else:
        log("All the required OpenGL functions are available: %s " % csv(available))

def get_max_texture_size() -> int:
    # pylint: disable=import-outside-toplevel
    from OpenGL.GL import glGetInteger, GL_MAX_TEXTURE_SIZE
    texture_size = glGetInteger(GL_MAX_TEXTURE_SIZE)
    log("GL_MAX_TEXTURE_SIZE=%s", texture_size)
    #this one may be missing?
    rect_texture_size = texture_size
    try:
        from OpenGL.GL import GL_MAX_RECTANGLE_TEXTURE_SIZE
        rect_texture_size = glGetInteger(GL_MAX_RECTANGLE_TEXTURE_SIZE)
    except ImportError as e:
        log("OpenGL: %s", e)
        log("using GL_MAX_TEXTURE_SIZE=%s as default", texture_size)
    except Exception as e:
        log("failed to query GL_MAX_RECTANGLE_TEXTURE_SIZE: %s", e)
    else:
        log("Texture size GL_MAX_RECTANGLE_TEXTURE_SIZE=%s", rect_texture_size)
    return int(min(rect_texture_size, texture_size))


def check_PyOpenGL_support(force_enable) -> Dict[str,Any]:
    from xpra.util import numpy_import_lock
    with numpy_import_lock:
        return do_check_PyOpenGL_support(force_enable)

def do_check_PyOpenGL_support(force_enable) -> Dict[str,Any]:
    props : Dict[str,Any] = {
        "platform"  : sys.platform,
        }
    def unsafe():
        props["safe"] = False
    # pylint: disable=import-outside-toplevel
    redirected_loggers: Dict[str, Tuple[Logger,List,bool]] = {}
    try:
        if CRASH:
            import ctypes
            ctypes.string_at(0)
            raise RuntimeError("should have crashed!")
        if TIMEOUT>0:
            import time
            time.sleep(TIMEOUT)
        #log redirection:
        for name in ("formathandler", "extensions", "acceleratesupport", "arrays", "converters"):
            logger = logging.getLogger(f"OpenGL.{name}")
            redirected_loggers[name] = (logger, list(logger.handlers), logger.propagate)
            logger.handlers = [CaptureHandler()]
            logger.propagate = False
        import OpenGL
        props["pyopengl"] = OpenGL.__version__  # @UndefinedVariable
        from OpenGL.GL import GL_VERSION, GL_EXTENSIONS
        from OpenGL.GL import glGetString, glGetIntegerv
        gl_version_str = glGetString(GL_VERSION)
        if gl_version_str is None and not force_enable:
            raise_fatal_error("OpenGL version is missing - cannot continue")
            return props
        #b'4.6.0 NVIDIA 440.59' -> ['4', '6', '0 NVIDIA...']
        log("GL_VERSION=%s", bytestostr(gl_version_str))
        vparts = bytestostr(gl_version_str).split(" ", 1)[0].split(".")
        try:
            gl_major = int(vparts[0])
            gl_minor = int(vparts[1])
        except (IndexError, ValueError) as e:
            log("failed to parse gl version '%s': %s", bytestostr(gl_version_str), e)
            log(" assuming this is at least 1.1 to continue")
            unsafe()
            gl_major = gl_minor = 0
        else:
            props["opengl"] = gl_major, gl_minor
            MIN_VERSION = (1,1)
            if (gl_major, gl_minor) < MIN_VERSION:
                if not force_enable:
                    raise_fatal_error("OpenGL output requires version %s or greater, not %s.%s" %
                                  (".".join([str(x) for x in MIN_VERSION]), gl_major, gl_minor))
                    return props
                unsafe()
            else:
                log("found valid OpenGL version: %s.%s", gl_major, gl_minor)

        from OpenGL import version as OpenGL_version
        pyopengl_version = OpenGL_version.__version__
        try:
            import OpenGL_accelerate            #@UnresolvedImport
            accel_version = OpenGL_accelerate.__version__
            props["accelerate"] = accel_version
            log("OpenGL_accelerate version %s", accel_version)
        except ImportError:
            log("OpenGL_accelerate not found")
            OpenGL_accelerate = None
            accel_version = None

        if accel_version is not None and pyopengl_version!=accel_version:
            global _version_warning_shown
            if not _version_warning_shown:
                log.warn("Warning: version mismatch between PyOpenGL and PyOpenGL-accelerate")
                log.warn(" %s vs %s", pyopengl_version, accel_version)
                log.warn(" this may cause crashes")
                _version_warning_shown = True
                gl_check_error(f"PyOpenGL vs accelerate version mismatch: {pyopengl_version} vs {accel_version}")
        vernum = parse_pyopengl_version(pyopengl_version)
        #we now require PyOpenGL 3.1.4 or later
        #3.1.4 was released in 2019
        if vernum<(3, 1, 4):
            if not force_enable:
                raise_fatal_error(f"PyOpenGL version {pyopengl_version} is too old and buggy")
                return {}
            unsafe()
        props["zerocopy"] = bool(OpenGL_accelerate) and pyopengl_version==accel_version

        try:
            extensions = glGetString(GL_EXTENSIONS).decode().split(" ")
            log("OpenGL extensions found: %s", csv(extensions))
            props["extensions"] = extensions
        except Exception:
            log("error querying extensions", exc_info=True)
            extensions = []
            if not force_enable:
                raise_fatal_error("OpenGL could not find the list of GL extensions -"+
                                  " does the graphics driver support OpenGL?")
            unsafe()

        from OpenGL.arrays.arraydatatype import ArrayDatatype
        try:
            log("found the following array handlers: %s", set(ArrayDatatype.getRegistry().values()))
        except Exception:
            pass

        from OpenGL.GL import GL_RENDERER, GL_VENDOR, GL_SHADING_LANGUAGE_VERSION
        def fixstring(v):
            try:
                return str(v).strip()
            except Exception:
                return str(v)
        for d,s,fatal in (("vendor",     GL_VENDOR,      True),
                          ("renderer",   GL_RENDERER,    True),
                          ("shading-language-version", GL_SHADING_LANGUAGE_VERSION, False)):
            try:
                v = glGetString(s)
                v = fixstring(v.decode())
                log("%s: %s", d, v)
            except Exception:
                if fatal and not force_enable:
                    gl_check_error(f"OpenGL property {d!r} is missing")
                else:
                    log(f"OpenGL property {d!r} is missing")
                v = ""
            props[d] = v
        vendor = props["vendor"]
        version_req = VERSION_REQ.get(vendor)
        if version_req:
            req_maj, req_min = version_req
            if gl_major<req_maj or (gl_major==req_maj and gl_minor<req_min):
                if force_enable:
                    log.warn("Warning: '%s' OpenGL driver requires version %i.%i", vendor, req_maj, req_min)
                    log.warn(" version %i.%i was found", gl_major, gl_minor)
                    unsafe()
                else:
                    gl_check_error("OpenGL version %i.%i is too old, %i.%i is required for %s" % (
                        gl_major, gl_minor, req_maj, req_min, vendor))

        from OpenGL.GLU import gluGetString, GLU_VERSION, GLU_EXTENSIONS
        #maybe we can continue without?
        if not bool(gluGetString):
            raise_fatal_error("no OpenGL GLU support")
        for d,s in {"GLU.version": GLU_VERSION, "GLU.extensions":GLU_EXTENSIONS}.items():
            v = gluGetString(s)
            v = v.decode()
            log("%s: %s", d, v)
            props[d] = v

        def match_list(thelist : GL_MATCH_LIST, listname:str):
            for k, values in thelist.items():
                prop = str(props.get(k))
                if prop and any(True for x in values if prop.find(x)>=0):
                    log("%s '%s' found in %s: %s", k, prop, listname, values)
                    return (k, prop)
                log("%s '%s' not found in %s: %s", k, prop, listname, values)
            return None
        blacklisted = match_list(BLACKLIST, "blacklist")
        greylisted = match_list(GREYLIST, "greylist")
        whitelisted = match_list(WHITELIST, "whitelist")
        if blacklisted:
            if whitelisted:
                log.info("%s '%s' enabled (found in both blacklist and whitelist)", *whitelisted)
            elif force_enable:
                log.warn("Warning: %s '%s' is blacklisted!", *blacklisted)
                log.warn(" force enabled by option")
            else:
                log.warn("%s '%s' is blacklisted!", *blacklisted)
                raise_fatal_error("%s '%s' is blacklisted!" % blacklisted)
        safe = bool(whitelisted) or not bool(blacklisted)
        if greylisted and not whitelisted:
            log.warn("Warning: %s '%s' is greylisted,", *greylisted)
            log.warn(" you may want to turn off OpenGL if you encounter bugs")
        if props.get("safe") is None:
            props["safe"] = safe

        #check for specific functions we need:
        from OpenGL.GL import (
            glActiveTexture, glTexSubImage2D, glTexCoord2i,
            glViewport, glMatrixMode, glLoadIdentity, glOrtho,
            glEnableClientState, glGenTextures, glDisable,
            glBindTexture, glPixelStorei, glEnable, glBegin, glFlush,
            glTexParameteri, glTexEnvi, glHint, glBlendFunc, glLineStipple,
            glTexImage2D,
            glMultiTexCoord2i,
            glVertex2i, glEnd,
            )
        check_functions(force_enable,
            glActiveTexture, glTexSubImage2D, glTexCoord2i,
            glViewport, glMatrixMode, glLoadIdentity, glOrtho,
            glEnableClientState, glGenTextures, glDisable,
            glBindTexture, glPixelStorei, glEnable, glBegin, glFlush,
            glTexParameteri, glTexEnvi, glHint, glBlendFunc, glLineStipple,
            glTexImage2D,
            glMultiTexCoord2i,
            glVertex2i, glEnd)
        #check for framebuffer functions we need:
        from OpenGL.GL.ARB.framebuffer_object import (
            GL_FRAMEBUFFER,
            GL_COLOR_ATTACHMENT0,
            glGenFramebuffers, glBindFramebuffer, glFramebufferTexture2D,
            )
        check_functions(force_enable,
            GL_FRAMEBUFFER,
            GL_COLOR_ATTACHMENT0,
            glGenFramebuffers, glBindFramebuffer, glFramebufferTexture2D,
            )

        try:
            from OpenGL.GL import glEnablei
        except ImportError:
            glEnablei = None
        global GL_ALPHA_SUPPORTED
        if not bool(glEnablei):
            log.warn("OpenGL glEnablei is not available, disabling transparency")
            GL_ALPHA_SUPPORTED = False
        props["transparency"] = GL_ALPHA_SUPPORTED

        missing_extensions = [ext for ext in required_extensions if ext not in extensions]
        if missing_extensions:
            if not force_enable:
                raise_fatal_error("OpenGL driver lacks support for extension: %s" % csv(missing_extensions))
            log("some extensions are missing: %s", csv(missing_extensions))
            unsafe()
        else:
            log("All required extensions are present: %s", required_extensions)

        #this allows us to do CSC via OpenGL:
        #see http://www.opengl.org/registry/specs/ARB/fragment_program.txt
        from OpenGL.GL.ARB.fragment_program import glInitFragmentProgramARB
        from OpenGL.GL.ARB.texture_rectangle import glInitTextureRectangleARB
        for name, fn in {
            "glInitFragmentProgramARB"  : glInitFragmentProgramARB,
            "glInitTextureRectangleARB" : glInitTextureRectangleARB,
            }.items():
            if not fn():
                if not force_enable:
                    raise_fatal_error("OpenGL output requires %s" % name)
                log("%s missing", name)
                unsafe()
            else:
                log("%s found", name)

        from OpenGL.GL.ARB.vertex_program import (
            glGenProgramsARB, glDeleteProgramsARB,
            glBindProgramARB, glProgramStringARB,
            )
        check_functions(force_enable,
                        glGenProgramsARB, glDeleteProgramsARB, glBindProgramARB, glProgramStringARB)

        texture_size_limit = get_max_texture_size()
        props["texture-size-limit"] = int(texture_size_limit)

        try:
            from OpenGL.GL import GL_MAX_VIEWPORT_DIMS
            v = glGetIntegerv(GL_MAX_VIEWPORT_DIMS)
            max_viewport_dims = int(v[0]), int(v[1])
            assert max_viewport_dims[0]>=texture_size_limit and max_viewport_dims[1]>=texture_size_limit
            log("GL_MAX_VIEWPORT_DIMS=%s", max_viewport_dims)
        except ImportError as e:
            log.error("Error querying max viewport dims: %s", e)
            max_viewport_dims = texture_size_limit, texture_size_limit
        props["max-viewport-dims"] = max_viewport_dims
        return props
    finally:
        def recs(name) -> List[str]:
            rlog = redirected_loggers.get(name)
            if not rlog:
                return []
            records = rlog[0].handlers[0].records
            return list(rec.getMessage() for rec in records)

        for msg in recs("acceleratesupport"):
            #strip default message prefix:
            msg = msg.replace("No OpenGL_accelerate module loaded: ", "")
            if msg=="No module named OpenGL_accelerate":
                msg = "missing accelerate module"
            if msg=="OpenGL_accelerate module loaded":
                log.info(msg)
            else:
                log.warn("PyOpenGL warning: %s", msg)

        #format handler messages:
        STRIP_LOG_MESSAGE = "Unable to load registered array format handler "
        missing_handlers = []
        for msg in recs("formathandler"):
            p = msg.find(STRIP_LOG_MESSAGE)
            if p<0:
                #unknown message, log it:
                log.info(msg)
                continue
            format_handler = msg[p+len(STRIP_LOG_MESSAGE):]
            p = format_handler.find(":")
            if p>0:
                format_handler = format_handler[:p]
                missing_handlers.append(format_handler)
        if missing_handlers:
            log.warn("PyOpenGL warning: missing array format handlers: %s", csv(missing_handlers))

        for msg in recs("extensions"):
            #ignore extension messages:
            p = msg.startswith("GL Extension ") and msg.endswith("available")
            if not p:
                log.info(msg)

        missing_accelerators = []
        STRIP_AR_HEAD = "Unable to load"
        STRIP_AR_TAIL = "from OpenGL_accelerate"
        for msg in recs("arrays")+recs("converters"):
            if msg.startswith(STRIP_AR_HEAD) and msg.endswith(STRIP_AR_TAIL):
                m = msg[len(STRIP_AR_HEAD):-len(STRIP_AR_TAIL)].strip()
                m = m.replace("accelerators", "").replace("accelerator", "").strip()
                missing_accelerators.append(m)
                continue
            if msg.startswith("Using accelerated"):
                log(msg)
            else:
                log.info(msg)
        if missing_accelerators:
            log.info("OpenGL accelerate missing: %s", csv(missing_accelerators))

        for logger, handlers, propagate in redirected_loggers.values():
            logger.handlers = handlers
            logger.propagate = propagate


def main() -> int:
    # pylint: disable=import-outside-toplevel
    from xpra.platform import program_context
    from xpra.platform.gui import init as gui_init
    from xpra.gtk_common.gtk_util import init_display_source
    from xpra.util import print_nested_dict
    from xpra.log import enable_color
    with program_context("OpenGL-Check"):
        gui_init()
        enable_color()
        verbose = "-v" in sys.argv or "--verbose" in sys.argv
        if verbose:
            log.enable_debug()
        init_display_source()
        force_enable = "-f" in sys.argv or "--force" in sys.argv
        from xpra.platform.gl_context import GLContext
        if not GLContext:
            log.warn("No OpenGL context implementation found")
            return 1
        log("testing %s", GLContext)
        gl_context = GLContext()  #pylint: disable=not-callable
        log("GLContext=%s", gl_context)
        #replace ImportError with a log message:
        global gl_check_error, gl_fatal_error
        errors = []
        def log_error(msg):
            log.error("ERROR: %s", msg)
            errors.append(msg)
        gl_check_error = log_error
        gl_fatal_error = log_error
        try:
            props = gl_context.check_support(force_enable)
        except Exception as e:
            props = {}
            log("check_support", exc_info=True)
            errors.append(e)
        log.info("")
        if errors:
            log.info("OpenGL errors:")
            for err in errors:
                log.info("  %s", err)
        if props:
            log.info("")
            log.info("OpenGL properties:")
            print_nested_dict(props)
        return len(errors)


if __name__ == "__main__":
    sys.exit(main())



#Copyright (c) 2018-2019 Jeremy Lainé.
#All rights reserved.
#
#Redistribution and use in source and binary forms, with or without
#modification, are permitted provided that the following conditions are met:
#
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright notice,
#      this list of conditions and the following disclaimer in the documentation
#      and/or other materials provided with the distribution.
#    * Neither the name of aioquic nor the names of its contributors may
#      be used to endorse or promote products derived from this software without
#      specific prior written permission.

#THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
#ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
#WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from typing import Dict, Optional
from aioquic.tls import SessionTicket


class SessionTicketStore:
    """
    Simple in-memory store for session tickets.
    """

    def __init__(self) -> None:
        self.tickets: Dict[bytes, SessionTicket] = {}

    def add(self, ticket: SessionTicket) -> None:
        self.tickets[ticket.ticket] = ticket

    def pop(self, label: bytes) -> Optional[SessionTicket]:
        return self.tickets.pop(label, None)



# -*- coding: utf-8 -*-
# This file is part of Xpra.
# Copyright (C) 2013-2023 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.


from weakref import WeakSet
from threading import Thread, Lock
from queue import Queue
from typing import Optional, Callable

from xpra.log import Logger
log = Logger("util")


class Worker_Thread(Thread):
    """
        A background thread which calls the functions we post to it.
        The functions are placed in a queue and only called once,
        when this thread gets around to it.
    """
    __slots__ = ("items", "exit", "daemon_work_items")

    def __init__(self):
        super().__init__(name="Worker_Thread", daemon=True)
        self.items : Queue[Optional[Callable]] = Queue()
        self.exit = False
        self.daemon_work_items : WeakSet[Callable] = WeakSet()

    def __repr__(self):
        return f"Worker_Thread(items={self.items.qsize()}, exit={self.exit})"

    def stop(self, force:bool=False) -> None:
        if self.exit:
            return
        items = tuple(x for x in self.items.queue if x is not None and x not in self.daemon_work_items)
        log("Worker_Thread.stop(%s) %i items still in work queue: %s", force, len(items), items)
        if force:
            if items:
                log.warn("Worker stop: %s items in the queue will not be run!", len(items))
                for x in list(self.items.queue):
                    if x:
                        log.warn(" - %s", x)
                self.items.put(None)
                self.items = Queue()
            self.exit = True
        else:
            if items:
                log.info("waiting for %s items in work queue to complete", len(items))
        self.items.put(None)

    def add(self, item:Callable, allow_duplicates:bool=True, daemon:bool=False) -> None:
        if self.items.qsize()>10:
            log.warn("Worker_Thread.items queue size is %s", self.items.qsize())
        if not allow_duplicates and item in self.items.queue:
            return
        self.items.put(item)
        if daemon:
            self.daemon_work_items.add(item)

    def run(self) -> None:
        log("Worker_Thread.run() starting")
        while not self.exit:
            item = self.items.get()
            if item is None:
                log("Worker_Thread.run() found end of queue marker")
                self.exit = True
                break
            try:
                log("Worker_Thread.run() calling %s (queue size=%s)", item, self.items.qsize())
                item()
            except Exception:
                log.error("Error in worker thread processing item %s", item, exc_info=True)
        log("Worker_Thread.run() ended (queue size=%s)", self.items.qsize())


# only one worker thread for now:
singleton : Optional[Worker_Thread]= None
# locking to ensure multithreaded code doesn't create more than one
lock = Lock()

def get_worker(create:bool=True) -> Optional[Worker_Thread]:
    global singleton
    #fast path (no lock):
    if singleton is not None or not create:
        return singleton
    with lock:
        if not singleton:
            singleton = Worker_Thread()
            singleton.start()
    return singleton

def add_work_item(item, allow_duplicates:bool=False, daemon:bool=True) -> None:
    w = get_worker(True)
    log("add_work_item(%s, %s, %s) worker=%s", item, allow_duplicates, daemon, w)
    assert w is not None
    w.add(item, allow_duplicates, daemon)

def stop_worker(force:bool=False) -> None:
    w = get_worker(False)
    log("stop_worker(%s) worker=%s", force, w)
    if w:
        w.stop(force)



# This file is part of Xpra.
# Copyright (C) 2008, 2009 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2013-2023 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import re
import sys
import binascii
import traceback
from itertools import chain
from threading import RLock
from enum import Enum, IntEnum
try:
    #Python 3.11 and later:
    from enum import StrEnum
except ImportError:
    StrEnum = Enum      # type: ignore
from typing import Dict, Tuple, Optional, Any, List, Set, Callable, Iterable, Union

# this is imported in a lot of places,
# so don't import too much at the top:
# pylint: disable=import-outside-toplevel

XPRA_APP_ID = 0

XPRA_GUID1 = 0x67b3efa2
XPRA_GUID2 = 0xe470
XPRA_GUID3 = 0x4a5f
XPRA_GUID4 = (0xb6, 0x53, 0x6f, 0x6f, 0x98, 0xfe, 0x60, 0x81)
XPRA_GUID_STR = "67B3EFA2-E470-4A5F-B653-6F6F98FE6081"
XPRA_GUID_BYTES = binascii.unhexlify(XPRA_GUID_STR.replace("-",""))


XPRA_NOTIFICATIONS_OFFSET = 2**24
class NotificationID(IntEnum):
    BANDWIDTH   = XPRA_NOTIFICATIONS_OFFSET+1
    IDLE        = XPRA_NOTIFICATIONS_OFFSET+2
    WEBCAM      = XPRA_NOTIFICATIONS_OFFSET+3
    AUDIO       = XPRA_NOTIFICATIONS_OFFSET+4
    OPENGL      = XPRA_NOTIFICATIONS_OFFSET+5
    SCALING     = XPRA_NOTIFICATIONS_OFFSET+6
    NEW_USER    = XPRA_NOTIFICATIONS_OFFSET+7
    CLIPBOARD   = XPRA_NOTIFICATIONS_OFFSET+8
    FAILURE     = XPRA_NOTIFICATIONS_OFFSET+9
    DPI         = XPRA_NOTIFICATIONS_OFFSET+10
    DISCONNECT  = XPRA_NOTIFICATIONS_OFFSET+11
    DISPLAY     = XPRA_NOTIFICATIONS_OFFSET+12
    STARTUP     = XPRA_NOTIFICATIONS_OFFSET+13
    FILETRANSFER    = XPRA_NOTIFICATIONS_OFFSET+14
    SHADOWWAYLAND   = XPRA_NOTIFICATIONS_OFFSET+15


#constants shared between client and server:
#(do not modify the values, see also disconnect_is_an_error)
#timeouts:
class ConnectionMessage(StrEnum):
    CLIENT_PING_TIMEOUT     = "client ping timeout"
    LOGIN_TIMEOUT           = "login timeout"
    CLIENT_EXIT_TIMEOUT     = "client exit timeout"
    #errors:
    PROTOCOL_ERROR          = "protocol error"
    VERSION_ERROR           = "version error"
    CONTROL_COMMAND_ERROR   = "control command error"
    AUTHENTICATION_FAILED   = "authentication failed"
    AUTHENTICATION_ERROR    = "authentication error"
    PERMISSION_ERROR        = "permission error"
    SERVER_ERROR            = "server error"
    CONNECTION_ERROR        = "connection error"
    SESSION_NOT_FOUND       = "session not found error"
    #informational (not a problem):
    DONE                    = "done"
    SERVER_EXIT             = "server exit"
    SERVER_UPGRADE          = "server upgrade"
    SERVER_SHUTDOWN         = "server shutdown"
    CLIENT_REQUEST          = "client request"
    DETACH_REQUEST          = "detach request"
    NEW_CLIENT              = "new client"
    IDLE_TIMEOUT            = "idle timeout"
    SESSION_BUSY            = "session busy"
    #client telling the server:
    CLIENT_EXIT             = "client exit"


#magic value for "workspace" window property, means unset
WORKSPACE_UNSET = 65535
WORKSPACE_ALL = 0xffffffff

WORKSPACE_NAMES = {
                   WORKSPACE_UNSET  : "unset",
                   WORKSPACE_ALL    : "all",
                   }

#this default value is based on 0.15.x clients,
#later clients should provide the `metadata.supported` capability instead
DEFAULT_METADATA_SUPPORTED = (
    "title", "icon-title", "pid", "iconic",
    "size-hints", "class-instance", "client-machine",
    "transient-for", "window-type",
    "fullscreen", "maximized", "decorations", "skip-taskbar", "skip-pager",
    "has-alpha", "override-redirect", "tray", "modal",
    "role", "opacity", "xid", "group-leader",
    "opaque-region",
    "command", "workspace", "above", "below", "sticky",
    "set-initial-position", "requested-position",
    "content-type",
    #4.4:
    #"parent", "relative-position",
    )


#initiate-moveresize X11 constants
class MoveResize(IntEnum):
    SIZE_TOPLEFT      = 0
    SIZE_TOP          = 1
    SIZE_TOPRIGHT     = 2
    SIZE_RIGHT        = 3
    SIZE_BOTTOMRIGHT  = 4
    SIZE_BOTTOM       = 5
    SIZE_BOTTOMLEFT   = 6
    SIZE_LEFT         = 7
    MOVE              = 8
    SIZE_KEYBOARD     = 9
    MOVE_KEYBOARD     = 10
    CANCEL            = 11
MOVERESIZE_DIRECTION_STRING = {
    MoveResize.SIZE_TOPLEFT      : "SIZE_TOPLEFT",
    MoveResize.SIZE_TOP          : "SIZE_TOP",
    MoveResize.SIZE_TOPRIGHT     : "SIZE_TOPRIGHT",
    MoveResize.SIZE_RIGHT        : "SIZE_RIGHT",
    MoveResize.SIZE_BOTTOMRIGHT  : "SIZE_BOTTOMRIGHT",
    MoveResize.SIZE_BOTTOM       : "SIZE_BOTTOM",
    MoveResize.SIZE_BOTTOMLEFT   : "SIZE_BOTTOMLEFT",
    MoveResize.SIZE_LEFT         : "SIZE_LEFT",
    MoveResize.MOVE              : "MOVE",
    MoveResize.SIZE_KEYBOARD     : "SIZE_KEYBOARD",
    MoveResize.MOVE_KEYBOARD     : "MOVE_KEYBOARD",
    MoveResize.CANCEL            : "CANCEL",
    }
SOURCE_INDICATION_UNSET     = 0
SOURCE_INDICATION_NORMAL    = 1
SOURCE_INDICATION_PAGER     = 2
SOURCE_INDICATION_STRING    = {
                               SOURCE_INDICATION_UNSET      : "UNSET",
                               SOURCE_INDICATION_NORMAL     : "NORMAL",
                               SOURCE_INDICATION_PAGER      : "PAGER",
                               }


util_logger = None
def get_util_logger():
    global util_logger
    if not util_logger:
        from xpra.log import Logger
        util_logger = Logger("util")
    return util_logger


#convenience method based on the strings above:
def disconnect_is_an_error(reason) -> bool:
    return reason.find("error")>=0 or (reason.find("timeout")>=0 and reason!=ConnectionMessage.IDLE_TIMEOUT)


def dump_exc():
    """Call this from an except: clause to print a nice traceback."""
    print("".join(traceback.format_exception(*sys.exc_info())))

def noerr(fn, *args):
    try:
        return fn(*args)
    except Exception:
        return None

def stderr_print(msg:str= "") -> bool:
    stderr = sys.stderr
    if stderr:
        try:
            noerr(stderr.write, msg+"\n")
            noerr(stderr.flush)
            return True
        except (OSError, AttributeError):
            pass
    return False


def nicestr(obj):
    """ Python 3.10 and older don't give us a nice string representation for enums """
    if isinstance(obj, Enum):
        return str(obj.value)
    return str(obj)


def net_utf8(value) -> str:
    """
    Given a value received by the network layer,
    convert it to a string.
    Gymnastics are involved if:
    - we get a memoryview from lz4
    - the rencode packet encoder is used
      as it ends up giving us a string which is actually utf8 bytes.
    """
    #with 'rencodeplus' or 'bencode', we just get the unicode string directly:
    if isinstance(value, str):
        return value
    if isinstance(value, memoryview):
        value = value.tobytes()
    #with rencode v1, we have to decode the value:
    #(after converting it to 'bytes' if necessary)
    return u(strtobytes(value))


def u(v) -> str:
    if isinstance(v, str):
        return v
    try:
        return v.decode("utf8")
    except (AttributeError, UnicodeDecodeError):
        return bytestostr(v)


# A simple little class whose instances we can stick random bags of attributes
# on.
class AdHocStruct:
    def __repr__(self):
        return ("<%s object, contents: %r>"
                % (type(self).__name__, self.__dict__))


def remove_dupes(seq:Iterable[Any]) -> List[Any]:
    seen : Set[Any] = set()
    seen_add : Callable = seen.add
    return [x for x in seq if not (x in seen or seen_add(x))]

def merge_dicts(a : Dict[str,Any], b : Dict[str,Any], path:Optional[List[str]]=None) -> Dict[str,Any]:
    """ merges b into a """
    if path is None:
        path = []
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                merge_dicts(a[key], b[key], path + [str(key)])
            elif a[key] == b[key]:
                pass # same leaf value
            else:
                raise ValueError('Conflict at %s: existing value is %s, new value is %s' % (
                    '.'.join(path + [str(key)]), a[key], b[key]))
        else:
            a[key] = b[key]
    return a

def make_instance(class_options, *args):
    log = get_util_logger()
    log("make_instance%s", tuple([class_options]+list(args)))
    for c in class_options:
        if c is None:
            continue
        try:
            v = c(*args)
            log(f"make_instance(..) {c}()={v}")
            if v:
                return v
        except Exception:
            log.error("make_instance(%s, %s)", class_options, args, exc_info=True)
            log.error("Error: cannot instantiate %s:", c)
            log.error(" with arguments %s", tuple(args))
    return None


def roundup(n : int, m : int) -> int:
    return (n + m - 1) & ~(m - 1)


class AtomicInteger:
    __slots__ = ("counter", "lock")
    def __init__(self, integer : int = 0):
        self.counter : int = integer
        from threading import RLock
        self.lock : RLock = RLock()

    def increase(self, inc = 1) -> int:
        with self.lock:
            self.counter = self.counter + inc
            return self.counter

    def decrease(self, dec = 1) -> int:
        with self.lock:
            self.counter = self.counter - dec
            return self.counter

    def get(self) -> int:
        return self.counter

    def __str__(self) -> str:
        return str(self.counter)

    def __repr__(self) -> str:
        return f"AtomicInteger({self.counter})"


    def __int__(self) -> int:
        return self.counter

    def __eq__(self, other) -> bool:
        try:
            return self.counter==int(other)
        except ValueError:
            return False

    def __cmp__(self, other) -> int:
        try:
            return self.counter-int(other)
        except ValueError:
            return -1


class MutableInteger(object):
    __slots__ = ("counter", )
    def __init__(self, integer : int = 0):
        self.counter : int = integer

    def increase(self, inc = 1) -> int:
        self.counter = self.counter + inc
        return self.counter

    def decrease(self, dec = 1) -> int:
        self.counter = self.counter - dec
        return self.counter

    def get(self) -> int:
        return self.counter

    def __str__(self) -> str:
        return str(self.counter)

    def __repr__(self) -> str:
        return f"MutableInteger({self.counter})"


    def __int__(self) -> int:
        return self.counter

    def __eq__(self, other) -> bool:
        return self.counter==int(other)
    def __ne__(self, other) -> bool:
        return self.counter!=int(other)
    def __lt__(self, other) -> bool:
        return self.counter<int(other)
    def __le__(self, other) -> bool:
        return self.counter<=int(other)
    def __gt__(self, other) -> bool:
        return self.counter>int(other)
    def __ge__(self, other) -> bool:
        return self.counter>=int(other)
    def __cmp__(self, other) -> int:
        return self.counter-int(other)


def strtobytes(x) -> bytes:
    if isinstance(x, bytes):
        return x
    return str(x).encode("latin1")
def bytestostr(x) -> str:
    if isinstance(x, (bytes, bytearray)):
        return x.decode("latin1")
    return str(x)

def decode_str(x, try_encoding="utf8"):
    """
    When we want to decode something (usually a byte string) no matter what.
    Try with utf8 first then fallback to just bytestostr().
    """
    try:
        return x.decode(try_encoding)
    except (AttributeError, UnicodeDecodeError):
        return bytestostr(x)


_RaiseKeyError = object()

def checkdict(v):
    assert isinstance(v, dict)
    return v

class typedict(dict):
    __slots__ = ("warn", ) # no __dict__ - that would be redundant
    @staticmethod # because this doesn't make sense as a global function.
    def _process_args(mapping=(), **kwargs) -> Dict[str,Any]:
        if hasattr(mapping, "items"):
            mapping = getattr(mapping, "items")()
        return dict((bytestostr(k), v) for k, v in chain(mapping, getattr(kwargs, "items")()))
    def __init__(self, mapping=(), **kwargs):
        super().__init__(self._process_args(mapping, **kwargs))
        self.warn = self._warn
    def __getitem__(self, k):
        return super().__getitem__(bytestostr(k))
    def __setitem__(self, k, v):
        return super().__setitem__(bytestostr(k), v)
    def __delitem__(self, k):
        return super().__delitem__(bytestostr(k))
    def get(self, k, default=None):
        kstr = bytestostr(k)
        if kstr in self:
            return super().get(kstr, default)
        #try to locate this value in a nested dictionary:
        if kstr.find(".")>0:
            prefix, k = kstr.split(".", 1)
            if prefix in self:
                v = super().get(prefix)
                if isinstance(v, dict):
                    return typedict(v).get(k, default)
        return default
    def setdefault(self, k, default=None):
        return super().setdefault(bytestostr(k), default)
    def pop(self, k, v=_RaiseKeyError):
        if v is _RaiseKeyError:
            return super().pop(bytestostr(k))
        return super().pop(bytestostr(k), v)
    def update(self, mapping=(), **kwargs):
        super().update(self._process_args(mapping, **kwargs))
    def __contains__(self, k):
        return super().__contains__(bytestostr(k))
    @classmethod
    def fromkeys(cls, keys, v=None):
        return super().fromkeys((bytestostr(k) for k in keys), v)
    def __repr__(self):
        return '{0}({1})'.format(type(self).__name__, super().__repr__())

    def _warn(self, msg, *args):
        get_util_logger().warn(msg, *args)

    def conv_get(self, k, default=None, conv=None):
        strkey = bytestostr(k)
        if strkey in self:
            v = super().get(strkey)
        else:
            #try harder by recursing:
            d = self
            while strkey.find(".")>0:
                prefix, k = strkey.split(".", 1)
                if prefix not in d:
                    return default
                v = d[prefix]
                if not isinstance(v, dict):
                    return default
                d = v
                strkey = k
            if strkey not in d:
                return default
            v = dict.get(d, strkey)
        if isinstance(v, dict) and conv and conv in (bytestostr, strtobytes, int, bool):
            d = typedict(v)
            if "" in d:
                v = d[""]
        try:
            return conv(v)
        except (TypeError, ValueError, AssertionError) as e:
            self._warn(f"Warning: failed to convert {k}")
            self._warn(f" from {type(v)} using {conv}: {e}")
            return default

    def uget(self, k, default=None):
        return self.conv_get(k, default, u)

    def strget(self, k, default:Optional[str]=None) -> str:
        return self.conv_get(k, default, bytestostr)

    def bytesget(self, k : str, default:Optional[bytes]=None) -> bytes:
        return self.conv_get(k, default, strtobytes)

    def intget(self, k : str, default:Optional[int]=0) -> int:
        return self.conv_get(k, default, int)

    def boolget(self, k : str, default:Optional[bool]=False) -> bool:
        return self.conv_get(k, default, bool)

    def dictget(self, k : str, default:Optional[dict]=None) -> Dict:
        return self.conv_get(k, default, checkdict)

    def intpair(self, k : str, default_value:Optional[Tuple[int,int]]=None) -> Optional[Tuple[int, int]]:
        v = self.inttupleget(k, default_value)
        if v is None:
            return default_value
        if len(v)!=2:
            #"%s is not a pair of numbers: %s" % (k, len(v))
            return default_value
        try:
            return int(v[0]), int(v[1])
        except ValueError:
            return default_value

    def strtupleget(self, k : str, default_value=(), min_items:Optional[int]=None, max_items:Optional[int]=None) -> Tuple[str, ...]:
        return self.tupleget(k, default_value, str, min_items, max_items)

    def inttupleget(self, k : str, default_value=(), min_items:Optional[int]=None, max_items:Optional[int]=None) -> Tuple[int, ...]:
        return self.tupleget(k, default_value, int, min_items, max_items)

    def tupleget(self, k : str, default_value=(), item_type=None, min_items:Optional[int]=None, max_items:Optional[int]=None) -> Tuple[Any, ...]:
        v = self._listget(k, default_value, item_type, min_items, max_items)
        return tuple(v or ())

    def _listget(self, k : str, default_value, item_type=None, min_items:Optional[int]=None, max_items:Optional[int]=None) -> List[Any]:
        v = self.get(k)
        if v is None:
            return default_value
        if isinstance(v, dict) and "" in v:
            v = v.get("")
        if not isinstance(v, (list, tuple)):
            self._warn("listget%s", (k, default_value, item_type, max_items))
            self._warn("expected a list or tuple value for %s but got %s", k, type(v))
            return default_value
        if min_items is not None and len(v)<min_items:
            self._warn("too few items in %s %s: minimum %s allowed, but got %s", type(v), k, max_items, len(v))
            return default_value
        if max_items is not None and len(v)>max_items:
            self._warn("too many items in %s %s: maximum %s allowed, but got %s", type(v), k, max_items, len(v))
            return default_value
        aslist = list(v)
        if item_type:
            for i, x in enumerate(aslist):
                if isinstance(x, bytes) and item_type==str:
                    x = bytestostr(x)
                    aslist[i] = x
                elif isinstance(x, str) and item_type==str:
                    x = str(x)
                    aslist[i] = x
                if not isinstance(x, item_type):
                    if callable(item_type):
                        try:
                            return item_type(x)
                        except Exception:
                            self._warn("invalid item type for %s %s: %s cannot be used with %s",
                                       type(v), k, item_type, type(x))
                            return default_value
                    self._warn("invalid item type for %s %s: expected %s but got %s",
                               type(v), k, item_type, type(x))
                    return default_value
        return aslist


def parse_scaling_value(v) -> Optional[Tuple[int,int]]:
    if not v:
        return None
    if v.endswith("%"):
        return float(v[:1]).as_integer_ratio()
    values = v.replace("/", ":").replace(",", ":").split(":", 1)
    values = [int(x) for x in values]
    for x in values:
        assert x>0, f"invalid scaling value {x}"
    if len(values)==1:
        ret = 1, values[0]
    else:
        assert values[0]<=values[1], "cannot upscale"
        ret = values[0], values[1]
    return ret

def from0to100(v):
    return intrangevalidator(v, 0, 100)

def intrangevalidator(v, min_value=None, max_value=None):
    v = int(v)
    if min_value is not None and v<min_value:
        raise ValueError(f"value must be greater than {min_value}")
    if max_value is not None and v>max_value:
        raise ValueError(f"value must be lower than {max_value}")
    return v


def log_screen_sizes(root_w, root_h, sizes):
    try:
        do_log_screen_sizes(root_w, root_h, sizes)
    except Exception as e:
        get_util_logger().warn("failed to parse screen size information: %s", e, exc_info=True)

def prettify_plug_name(s, default="") -> str:
    if not s:
        return default
    try:
        s = s.decode("utf8")
    except (AttributeError, UnicodeDecodeError):
        pass
    #prettify strings on win32
    s = re.sub(r"[0-9\.]*\\", "-", s).lstrip("-")
    if s.startswith("WinSta-"):
        s = s[len("WinSta-"):]
    #ie: "(Standard monitor types) DELL ..."
    if s.startswith("(") and s.lower().find("standard")<s.find(") "):
        s = s.split(") ", 1)[1]
    if s=="0":
        s = default
    return s

def do_log_screen_sizes(root_w, root_h, sizes):
    from xpra.log import Logger
    log = Logger("screen")
    #old format, used by some clients (android):
    if not isinstance(sizes, (tuple, list)):
        return
    if any(True for x in sizes if not isinstance(x, (tuple, list))):
        return
    def dpi(size_pixels, size_mm):
        if size_mm==0:
            return 0
        return round(size_pixels * 254 / size_mm / 10)
    def add_workarea(info, wx, wy, ww, wh):
        info.append("workarea: %4ix%-4i" % (ww, wh))
        if wx!=0 or wy!=0:
            #log position if not (0, 0)
            info.append("at %4ix%-4i" % (wx, wy))
    if len(sizes)!=1:
        log.warn("Warning: more than one screen found")
        log.warn(" this is not supported")
        log("do_log_screen_sizes(%i, %i, %s)", root_w, root_h, sizes)
        return
    s = sizes[0]
    if len(s)<10:
        log.info(" %s", s)
        return
    #more detailed output:
    display_name, width, height, width_mm, height_mm, \
    monitors, work_x, work_y, work_width, work_height = s[:10]
    #always log plug name:
    info = ["%s" % prettify_plug_name(display_name)]
    if width!=root_w or height!=root_h:
        #log plug dimensions if not the same as display (root):
        info.append("%ix%i" % (width, height))
    sdpix = dpi(width, width_mm)
    sdpiy = dpi(height, height_mm)
    info.append("(%ix%i mm - DPI: %ix%i)" % (width_mm, height_mm, sdpix, sdpiy))

    if work_width!=width or work_height!=height or work_x!=0 or work_y!=0:
        add_workarea(info, work_x, work_y, work_width, work_height)
    log.info("  "+" ".join(info))
    #sort monitors from left to right, top to bottom:
    monitors_distances = []
    for m in monitors:
        plug_x, plug_y = m[1:3]
        monitors_distances.append((plug_x+plug_y*width, m))
    sorted_monitors = [x[1] for x in sorted(monitors_distances)]
    for i, m in enumerate(sorted_monitors, start=1):
        if len(m)<7:
            log.info("    %s", m)
            continue
        plug_name, plug_x, plug_y, plug_width, plug_height, plug_width_mm, plug_height_mm = m[:7]
        default_name = "monitor %i" % i
        info = ['%-16s' % prettify_plug_name(plug_name, default_name)]
        if plug_width!=width or plug_height!=height or plug_x!=0 or plug_y!=0:
            info.append("%4ix%-4i" % (plug_width, plug_height))
            if plug_x!=0 or plug_y!=0 or len(sorted_monitors)>1:
                info.append("at %4ix%-4i" % (plug_x, plug_y))
        if (plug_width_mm!=width_mm or plug_height_mm!=height_mm) and (plug_width_mm>0 or plug_height_mm>0):
            dpix = dpi(plug_width, plug_width_mm)
            dpiy = dpi(plug_height, plug_height_mm)
            dpistr = ""
            if sdpix!=dpix or sdpiy!=dpiy or len(sorted_monitors)>1:
                dpistr = " - DPI: %ix%i" % (dpix, dpiy)
            info.append("(%3ix%-3i mm%s)" % (plug_width_mm, plug_height_mm, dpistr))
        if len(m)>=11:
            dwork_x, dwork_y, dwork_width, dwork_height = m[7:11]
            #only show it again if different from the screen workarea
            if dwork_x!=work_x or dwork_y!=work_y or dwork_width!=work_width or dwork_height!=work_height:
                add_workarea(info, dwork_x, dwork_y, dwork_width, dwork_height)
        if len(sorted_monitors)==1 and len(info)==1 and info[0].strip() in ("Canvas", "DUMMY0"):
            #no point in logging just `Canvas` on its own
            continue
        istr = (" ".join(info)).rstrip(" ")
        if len(monitors)==1 and istr.lower() in ("unknown unknown", "0", "1", default_name, "screen", "monitor"):
            #a single monitor with no real name,
            #so don't bother showing it:
            continue
        log.info("    "+istr)

def get_screen_info(screen_sizes) -> Dict[int, Dict[str, Any]]:
    #same format as above
    if not screen_sizes:
        return {}
    info : Dict[int, Dict[str, Any]] = {}
    for i, x in enumerate(screen_sizes):
        if not isinstance(x, (tuple, list)):
            continue
        sinfo : Dict[str,Any] = info.setdefault(i, {})
        sinfo["display"] = x[0]
        if len(x)>=3:
            sinfo["size"] = x[1], x[2]
        if len(x)>=5:
            sinfo["size_mm"] = x[3], x[4]
        if len(x)>=6:
            monitors = x[5]
            for j, monitor in enumerate(monitors):
                if len(monitor)>=7:
                    minfo : Dict[str,Any] = sinfo.setdefault("monitor", {}).setdefault(j, {})
                    for k,v in {
                                "name"      : monitor[0],
                                "geometry"  : monitor[1:5],
                                "size_mm"   : monitor[5:7],
                                }.items():
                        minfo[k] = v
        if len(x)>=10:
            sinfo["workarea"] = x[6:10]
    return info

def dump_all_frames(logger=None) -> None:
    try:
        frames = sys._current_frames()      #pylint: disable=protected-access
    except AttributeError:
        return
    else:
        dump_frames(frames.items(), logger)

def dump_gc_frames(logger=None) -> None:
    import gc
    import inspect
    gc.collect()
    frames = tuple((None, x) for x in gc.get_objects() if inspect.isframe(x))
    dump_frames(frames, logger)

def dump_frames(frames, logger=None) -> None:
    if not logger:
        logger = get_util_logger()
    logger("found %s frames:", len(frames))
    for i,(fid,frame) in enumerate(frames):
        fidstr = ""
        if fid is not None:
            try:
                fidstr = hex(fid)
            except TypeError:
                fidstr = str(fid)
        logger("%i: %s %s:", i, fidstr, frame)
        for x in traceback.format_stack(frame):
            for l in x.splitlines():
                logger("%s", l)


def detect_leaks() -> Callable[[], None]:
    import tracemalloc
    tracemalloc.start()
    last_snapshot = [tracemalloc.take_snapshot()]
    def print_leaks():
        s1 = last_snapshot[0]
        s2 = tracemalloc.take_snapshot()
        last_snapshot[0] = s2
        top_stats = s2.compare_to(s1, 'lineno')
        print("[ Top 20 differences ]")
        for stat in top_stats[:20]:
            print(stat)
        for i, stat in enumerate(top_stats[:20]):
            print()
            print("top %i:" % i)
            print("%s memory blocks: %.1f KiB" % (stat.count, stat.size / 1024))
            for line in stat.traceback.format():
                print(line)
        return True
    return print_leaks

def start_mem_watcher(ms) -> None:
    from xpra.make_thread import start_thread
    start_thread(mem_watcher, name="mem-watcher", daemon=True, args=(ms,))

def mem_watcher(ms, pid:int=os.getpid()) -> None:
    import time
    import psutil
    process = psutil.Process(pid)
    while True:
        mem = process.memory_full_info()
        #get_util_logger().info("memory usage: %s", mem.mem//1024//1024)
        get_util_logger().info("memory usage for %s: %s", pid, mem)
        time.sleep(ms/1000.0)

def log_mem_info(prefix="memory usage: ", pid=os.getpid()) -> None:
    import psutil
    process = psutil.Process(pid)
    mem = process.memory_full_info()
    print("%i %s%s" % (pid, prefix, mem))


class ellipsizer:
    __slots__ = ("obj", "limit")
    def __init__(self, obj, limit=100):
        self.obj = obj
        self.limit = limit
    def __str__(self):
        return self.__repr__()
    def __repr__(self):
        if self.obj is None:
            return "None"
        return repr_ellipsized(self.obj, self.limit)

def repr_ellipsized(obj, limit=100) -> str:
    if isinstance(obj, str):
        if len(obj)>limit>6:
            return nonl(obj[:limit//2-2]+" .. "+obj[2-limit//2:])
        return nonl(obj)
    if isinstance(obj, memoryview):
        obj = obj.tobytes()
    if isinstance(obj, bytes):
        try:
            s = nonl(repr(obj))
        except Exception:
            s = binascii.hexlify(obj).decode()
        if len(s)>limit>6:
            return nonl(s[:limit//2-2]+" .. "+s[2-limit//2:])
        return s
    return repr_ellipsized(repr(obj), limit)


def rindex(alist:Union[List,Tuple], avalue:Any) -> int:
    return len(alist) - alist[::-1].index(avalue) - 1


def notypedict(d:Dict) -> Dict:
    for k in list(d.keys()):
        v = d[k]
        if isinstance(v, dict):
            d[k] = notypedict(v)
    return dict(d)

def flatten_dict(info:Dict[str,Any], sep:str=".") -> Dict[str,Any]:
    to : Dict[str, Any] = {}
    _flatten_dict(to, sep, "", info)
    return to

def _flatten_dict(to:Dict[str, Any], sep:str, path:str, d:Dict[str,Any]):
    for k,v in d.items():
        if path:
            if k:
                npath = path+sep+bytestostr(k)
            else:
                npath = path
        else:
            npath = bytestostr(k)
        if isinstance(v, dict):
            _flatten_dict(to, sep, npath, v)
        elif v is not None:
            to[npath] = v

def parse_simple_dict(s:str="", sep:str=",") -> Dict[str, Union[str,List[str]]]:
    #parse the options string and add the pairs:
    d : Dict[str, Union[str,List[str]]] = {}
    for el in s.split(sep):
        if not el:
            continue
        try:
            k, v = el.split("=", 1)
            def may_add() -> Union[str,List[str]]:
                cur = d.get(k)
                if cur is None:
                    return v
                if not isinstance(cur, list):
                    cur = [cur]
                cur.append(v)
                return cur
            d[k] = may_add()
        except Exception as e:
            log = get_util_logger()
            log.warn("Warning: failed to parse dictionary option '%s':", s)
            log.warn(" %s", e)
    return d

#used for merging dicts with a prefix and suffix
#non-None values get added to <todict> with a prefix and optional suffix
def updict(todict:Dict, prefix:str, d:Dict, suffix:str="", flatten_dicts:bool=False) -> Dict:
    if not d:
        return todict
    for k,v in d.items():
        if v is not None:
            if k:
                k = prefix+"."+str(k)
            else:
                k = prefix
            if suffix:
                k = k+"."+suffix
            if flatten_dicts and isinstance(v, dict):
                updict(todict, k, v)
            else:
                todict[k] = v
    return todict

def pver(v, numsep:str=".", strsep:str=", ") -> str:
    #print for lists with version numbers, or CSV strings
    if isinstance(v, (list, tuple)):
        types = list(set(type(x) for x in v))
        if len(types)==1:
            if types[0]==int:
                return numsep.join(str(x) for x in v)
            if types[0]==str:
                return strsep.join(str(x) for x in v)
            if types[0]==bytes:
                def s(x):
                    try:
                        return x.decode("utf8")
                    except UnicodeDecodeError:
                        return bytestostr(x)
                return strsep.join(s(x) for x in v)
    return bytestostr(v)

def sorted_nicely(l:Iterable):
    """ Sort the given iterable in the way that humans expect."""
    def convert(text):
        if text.isdigit():
            return int(text)
        return text
    alphanum_key = lambda key: [convert(c) for c in re.split(r"(\d+)", bytestostr(key))]
    return sorted(l, key = alphanum_key)

def print_nested_dict(d:Dict, prefix:str="", lchar:str="*", pad:int=32, vformat=None, print_fn:Optional[Callable]=None,
                      version_keys=("version", "revision"), hex_keys=("data", )):
    #"smart" value formatting function:
    def sprint(arg):
        if print_fn:
            print_fn(arg)
        else:
            print(arg)
    def vf(k, v):
        if vformat:
            fmt = vformat
            if isinstance(vformat, dict):
                fmt = vformat.get(k)
            if fmt is not None:
                return nonl(fmt(v))
        try:
            if any(k.find(x)>=0 for x in version_keys):
                return nonl(pver(v)).lstrip("v")
            if any(k.find(x)>=0 for x in hex_keys):
                return binascii.hexlify(v)
        except Exception:
            pass
        return nonl(pver(v, ", ", ", "))
    l = pad-len(prefix)-len(lchar)
    for k in sorted_nicely(d.keys()):
        v = d[k]
        if isinstance(v, dict):
            nokey = v.get("", (v.get(None)))
            if nokey is not None:
                sprint("%s%s %s : %s" % (prefix, lchar, bytestostr(k).ljust(l), vf(k, nokey)))
                for x in ("", None):
                    v.pop(x, None)
            else:
                sprint("%s%s %s" % (prefix, lchar, bytestostr(k)))
            print_nested_dict(v, prefix+"  ", "-", vformat=vformat, print_fn=print_fn,
                              version_keys=version_keys, hex_keys=hex_keys)
        else:
            sprint("%s%s %s : %s" % (prefix, lchar, bytestostr(k).ljust(l), vf(k, v)))

def reverse_dict(d:Dict) -> Dict:
    reversed_d = {}
    for k,v in d.items():
        reversed_d[v] = k
    return reversed_d


def std(s, extras:str="-,./: ") -> str:
    s = s or ""
    try:
        s = s.decode("latin1")
    except Exception:
        pass
    def c(v):
        try:
            return chr(v)
        except Exception:
            return str(v)
    def f(v):
        return str.isalnum(c(v)) or v in extras
    return "".join(filter(f, s))

def alnum(s) -> str:
    try:
        s = s.encode("latin1")
    except Exception:
        pass
    def c(v):
        try:
            return chr(v)
        except Exception:
            return str(v)
    def f(v):
        return str.isalnum(c(v))
    return "".join(c(v) for v in filter(f, s))

def nonl(x) -> str:
    if not x:
        return ""
    return str(x).replace("\n", "\\n").replace("\r", "\\r")

def engs(v) -> str:
    if isinstance(v, int):
        l = v
    else:
        try:
            l = len(v)
        except TypeError:
            return ""
    return "s" if l!=1 else ""


def obsc(v) -> str:
    OBSCURE_PASSWORDS = envbool("XPRA_OBSCURE_PASSWORDS", True)
    if OBSCURE_PASSWORDS:
        return "".join("*" for _ in (bytestostr(v) or ""))
    return v


def csv(v) -> str:
    try:
        return ", ".join(str(x) for x in v)
    except Exception:
        return str(v)


def unsetenv(*varnames) -> None:
    for x in varnames:
        os.environ.pop(x, None)

def hasenv(name : str) -> bool:
    return os.environ.get(name) is not None

def envint(name : str, d:int=0) -> int:
    try:
        return int(os.environ.get(name, d))
    except ValueError:
        return d

def envbool(name : str, d:bool=False) -> bool:
    try:
        v = os.environ.get(name, "").lower()
        if v is None:
            return d
        if v in ("yes", "true", "on"):
            return True
        if v in ("no", "false", "off"):
            return False
        return bool(int(v))
    except ValueError:
        return d

def envfloat(name : str, d:float=0) -> float:
    try:
        return float(os.environ.get(name, d))
    except ValueError:
        return d


#give warning message just once per key then ignore:
_once_only = set()
def first_time(key:str) -> bool:
    if key not in _once_only:
        _once_only.add(key)
        return True
    return False

numpy_import_lock = RLock()




# This file is part of Xpra.
# Copyright (C) 2011-2023 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import hmac
import hashlib
from typing import List, Optional, Callable, ByteString

from xpra.util import csv, envint
from xpra.log import Logger
from xpra.os_util import strtobytes, memoryview_to_bytes, hexstr

log = Logger("network", "crypto")

BLACKLISTED_HASHES = ("sha1", "md5")
DEFAULT_SALT_LENGTH = envint("XPRA_DEFAULT_SALT_LENGTH", 64)


def get_digests() -> List[str]:
    digests = ["xor"]
    digests += [f"hmac+{x}" for x in tuple(reversed(sorted(hashlib.algorithms_available)))
                if not x.startswith("shake_") and x not in BLACKLISTED_HASHES
                and getattr(hashlib, x, None) is not None]
    try:
        from xpra.net.rfb import d3des  #pylint: disable=import-outside-toplevel
        assert d3des
        digests.append("des")
    except (ImportError, TypeError):    # pragma: no cover
        pass
    return digests

def get_digest_module(digest : str) -> Optional[Callable]:
    log(f"get_digest_module({digest})")
    if not digest or not digest.startswith("hmac"):
        return None
    try:
        digest_module = digest.split("+")[1]        #ie: "hmac+sha512" -> "sha512"
    except IndexError:
        return None
    try:
        return getattr(hashlib, digest_module)
    except AttributeError as e:
        log("no '%s' attribute in hashlib: %s", digest_module, e)
        return None

def choose_digest(options) -> str:
    assert len(options)>0, "no digest options"
    log(f"choose_digest({options})")
    #prefer stronger hashes:
    for h in ("sha512", "sha384", "sha256", "sha224"):
        hname = f"hmac+{h}"
        if hname in options:
            return hname
    if "xor" in options:
        return "xor"
    if "des" in options:
        return "des"
    raise ValueError(f"no known digest options found in '{csv(options)}'")

def gendigest(digest:str, password_in, salt_in:ByteString) -> bytes:
    assert password_in and salt_in
    salt : bytes = memoryview_to_bytes(salt_in)
    password : bytes = strtobytes(password_in)
    if digest=="des":
        from xpra.net.rfb.d3des import generate_response  #pylint: disable=import-outside-toplevel
        password = password.ljust(8, b"\x00")[:8]
        salt = salt.ljust(16, b"\x00")[:16]
        v = generate_response(password, salt)
        return strtobytes(hexstr(v))
    if digest in ("xor", "kerberos", "gss", "keycloak"):
        #kerberos, gss and keycloak use xor because we need to use the actual token
        #at the other end
        salt = salt.ljust(len(password), b"\x00")[:len(password)]
        from xpra.buffers.cyxor import xor_str           #@UnresolvedImport pylint: disable=import-outside-toplevel
        v = xor_str(password, salt)
        return memoryview_to_bytes(v)
    digestmod = get_digest_module(digest)
    if not digestmod:
        log(f"invalid digest module {digest!r}")
        return b""
        #warn_server_and_exit(ExitCode.UNSUPPORTED,
        #    "server requested digest '%s' but it is not supported" % digest, "invalid digest")
    return strtobytes(hmac.HMAC(password, salt, digestmod=digestmod).hexdigest())

def verify_digest(digest:str, password:str, salt, challenge_response:bytes) -> bool:
    if not password or not salt or not challenge_response:
        return False
    verify = gendigest(digest, password, salt)
    if not hmac.compare_digest(verify, challenge_response):
        log(f"expected {verify!r} but got {challenge_response!r}")
        return False
    return True


def get_salt(l:int=DEFAULT_SALT_LENGTH) -> bytes:
    #too short: we would not feed enough random data to HMAC
    if l<32:
        raise ValueError(f"salt is too short: only {l} bytes")
    #too long: limit the amount of random data we request from the system
    if l>=1024:
        raise ValueError(f"salt is too long: {l} bytes")
    #all server versions support a client salt,
    #they also tell us which digest to use:
    return os.urandom(l)



# This file is part of Xpra.
# Copyright (C) 2023 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.


#ctypedef int cudaError_t
#cdef extern from "cuda_runtime_api.h":
#    const char *cudaGetErrorName(cudaError_t err)
#    const char *cudaGetErrorString(cudaError_t err)

def cudacheck(r, fn=None):
    if r:
        msg = get_error_name(r)
        if fn:
            msg = fn+f": {msg!r}"
        raise RuntimeError(msg)


def get_error_name(error):
    return CUDA_ERRORS.get(error) or str(error)

CUDA_ERRORS = {
    0   : "SUCCESS",
    1   : "INVALID_VALUE",
    2   : "OUT_OF_MEMORY",
    3   : "NOT_INITIALIZED",
    4   : "DEINITIALIZED",
    5   : "PROFILER_DISABLED",
    6   : "PROFILER_NOT_INITIALIZED",
    7   : "PROFILER_ALREADY_STARTED",
    8   : "PROFILER_ALREADY_STOPPED",
    34  : "STUB_LIBRARY",
    46  : "DEVICE_UNAVAILABLE",
    100 : "NO_DEVICE",
    101 : "INVALID_DEVICE",
    102 : "DEVICE_NOT_LICENSED",
    200 : "INVALID_IMAGE",
    201 : "INVALID_CONTEXT",
    202 : "CONTEXT_ALREADY_CURRENT",
    205 : "MAP_FAILED",
    206 : "UNMAP_FAILED",
    207 : "ARRAY_IS_MAPPED",
    208 : "ALREADY_MAPPED",
    209 : "NO_BINARY_FOR_GPU",
    210 : "ALREADY_ACQUIRED",
    211 : "NOT_MAPPED",
    212 : "NOT_MAPPED_AS_ARRAY",
    213 : "NOT_MAPPED_AS_POINTER",
    214 : "ECC_UNCORRECTABLE",
    215 : "UNSUPPORTED_LIMIT",
    216 : "CONTEXT_ALREADY_IN_USE",
    217 : "PEER_ACCESS_UNSUPPORTED",
    218 : "INVALID_PTX",
    219 : "INVALID_GRAPHICS_CONTEXT",
    220 : "NVLINK_UNCORRECTABLE",
    221 : "JIT_COMPILER_NOT_FOUND",
    222 : "UNSUPPORTED_PTX_VERSION",
    223 : "JIT_COMPILATION_DISABLED",
    224 : "UNSUPPORTED_EXEC_AFFINITY",
    300 : "INVALID_SOURCE",
    301 : "FILE_NOT_FOUND",
    302 : "SHARED_OBJECT_SYMBOL_NOT_FOUND",
    303 : "SHARED_OBJECT_INIT_FAILED",
    304 : "OPERATING_SYSTEM",
    400 : "INVALID_HANDLE",
    401 : "ILLEGAL_STATE",
    500 : "NOT_FOUND",
    600 : "NOT_READY",
    700 : "ILLEGAL_ADDRESS",
    701 : "LAUNCH_OUT_OF_RESOURCES",
    702 : "LAUNCH_TIMEOUT",
    703 : "LAUNCH_INCOMPATIBLE_TEXTURING",
    704 : "PEER_ACCESS_ALREADY_ENABLED",
    705 : "PEER_ACCESS_NOT_ENABLED",
    708 : "PRIMARY_CONTEXT_ACTIVE",
    709 : "CONTEXT_IS_DESTROYED",
    710 : "ASSERT",
    711 : "TOO_MANY_PEERS",
    712 : "HOST_MEMORY_ALREADY_REGISTERED",
    713 : "HOST_MEMORY_NOT_REGISTERED",
    714 : "HARDWARE_STACK_ERROR",
    715 : "ILLEGAL_INSTRUCTION",
    716 : "MISALIGNED_ADDRESS",
    717 : "INVALID_ADDRESS_SPACE",
    718 : "INVALID_PC",
    719 : "LAUNCH_FAILED",
    720 : "COOPERATIVE_LAUNCH_TOO_LARGE",
    800 : "NOT_PERMITTED",
    801 : "NOT_SUPPORTED",
    999 : "UNKNOWN",
    802 : "SYSTEM_NOT_READY",
    803 : "SYSTEM_DRIVER_MISMATCH",
    804 : "COMPAT_NOT_SUPPORTED_ON_DEVICE",
    805 : "MPS_CONNECTION_FAILED",
    806 : "MPS_RPC_FAILURE",
    807 : "MPS_SERVER_NOT_READY",
    808 : "MPS_MAX_CLIENTS_REACHED",
    809 : "MPS_MAX_CONNECTIONS_REACHED",
    810 : "MPS_CLIENT_TERMINATED",
    811 : "CDP_NOT_SUPPORTED",
    812 : "CDP_VERSION_MISMATCH",
    900 : "STREAM_CAPTURE_UNSUPPORTED",
    901 : "STREAM_CAPTURE_INVALIDATED",
    902 : "STREAM_CAPTURE_MERGE",
    903 : "STREAM_CAPTURE_UNMATCHED",
    904 : "STREAM_CAPTURE_UNJOINED",
    905 : "STREAM_CAPTURE_ISOLATION",
    906 : "STREAM_CAPTURE_IMPLICIT",
    907 : "CAPTURED_EVENT",
    908 : "STREAM_CAPTURE_WRONG_THREAD",
    909 : "TIMEOUT",
    910 : "GRAPH_EXEC_UPDATE_FAILURE",
    911 : "EXTERNAL_DEVICE",
    912 : "INVALID_CLUSTER_SIZE",
    }



# This file is part of Xpra.
# Copyright (C) 2014-2023 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import sys
import subprocess
from typing import Optional, Callable

from xpra.platform.win32 import constants as win32con
from xpra.util import csv, envint, envbool, reverse_dict
from xpra.os_util import bytestostr
from xpra.platform.paths import get_app_dir
from xpra.log import Logger

log = Logger("printing")

#allows us to skip some printers we don't want to export
WINSPOOL_LISTENER = envbool("XPRA_WINSPOOL_LISTENER", True)
SKIPPED_PRINTERS = os.environ.get("XPRA_SKIPPED_PRINTERS", "Microsoft XPS Document Writer,Fax").split(",")
PRINTER_LEVEL = envint("XPRA_WIN32_PRINTER_LEVEL", 1)
DEFAULT_PRINTER_FLAGS = "LOCAL,SHARED+NETWORK+CONNECTIONS"
PRINTER_FLAGS = [x.strip() for x in os.environ.get("XPRA_WIN32_PRINTER_FLAGS", DEFAULT_PRINTER_FLAGS).split(",")]

DEFAULT_MIMETYPES = ["application/pdf", ]


PRINTER_ENUM_VALUES = {
    "DEFAULT"       : 1,
    "LOCAL"         : 2,
    "CONNECTIONS"   : 4,
    "NAME"          : 8,
    "REMOTE"        : 16,
    "SHARED"        : 32,
    "NETWORK"       : 64,
    "EXPAND"        : 16384,
    "CONTAINER"     : 32768,
    "ICON1"         : 65536*1,
    "ICON2"         : 65536*2,
    "ICON3"         : 65536*3,
    "ICON4"         : 65536*4,
    "ICON5"         : 65536*5,
    "ICON6"         : 65536*6,
    "ICON7"         : 65536*7,
    "ICON8"         : 65536*8,
    }
PRINTER_ENUM_NAMES = reverse_dict(PRINTER_ENUM_VALUES)
log("PRINTER_ENUM_VALUES: %s", PRINTER_ENUM_VALUES)

log("PRINTER_FLAGS=%s", csv(PRINTER_FLAGS))
VALID_PRINTER_FLAGS = ("LOCAL", "SHARED", "CONNECTIONS", "NETWORK", "REMOTE")
def get_printer_enums():
    printer_enums = []
    for v in PRINTER_FLAGS:                     #ie: "SHARED+NETWORK+CONNECTIONS"
        flags = v.replace('|','+').split("+")   #ie: ["SHARED", "NETWORK", "CONNECTIONS"]
        values = []
        for flag in flags:                      #ie: "SHARED"
            if flag not in VALID_PRINTER_FLAGS:
                log.warn("Warning: the following printer flag is invalid and will be ignored: %s", flag)
            else:
                values.append(flag)             #ie: "SHARED"
        printer_enums.append(values)
    return tuple(printer_enums)
PRINTER_ENUMS = get_printer_enums()
log("PRINTER_ENUMS=%s", PRINTER_ENUMS)


#emulate pycups job id
JOB_ID = 0
PROCESSES = {}

printers_modified_callback : Optional[Callable] = None
def init_printing(callback=None):
    global printers_modified_callback
    log("init_printing(%s) printers_modified_callback=%s", callback, printers_modified_callback)
    printers_modified_callback = callback
    if WINSPOOL_LISTENER:
        try:
            init_winspool_listener()
        except Exception:
            log.error("Error: failed to register for print spooler changes", exc_info=True)

def init_winspool_listener():
    from xpra.platform.win32.win32_events import get_win32_event_listener
    get_win32_event_listener().add_event_callback(win32con.WM_DEVMODECHANGE, on_devmodechange)

def on_devmodechange(wParam, lParam):
    global printers_modified_callback
    log("on_devmodechange(%s, %s) printers_modified_callback=%s", wParam, lParam, printers_modified_callback)
    #from ctypes import c_wchar_p
    #name = c_wchar_p(lParam)
    #log("device changed: %s", name)
    if lParam>0 and printers_modified_callback:
        printers_modified_callback()


def EnumPrinters(flags, name=None, level=PRINTER_LEVEL):
    from ctypes import (
        WinDLL,  # @UnresolvedImport
        byref, addressof, cast, Structure, POINTER,
        )
    from ctypes.wintypes import BYTE, BOOL, DWORD, LPCWSTR, LPBYTE, LPDWORD

    winspool = WinDLL('winspool.drv', use_last_error=True)
    EnumPrintersW = winspool.EnumPrintersW
    EnumPrintersW.restype = BOOL
    EnumPrintersW.argtypes = [DWORD, LPCWSTR, DWORD, LPBYTE, DWORD, LPDWORD, LPDWORD]

    class PRINTER_INFO(Structure):
        _fields_ = [
            ("Flags", DWORD),
            ("pDescription", LPCWSTR),
            ("pName", LPCWSTR),
            ("pComment", LPCWSTR),
        ]

    # Invoke once with a NULL pointer to get buffer size.
    info = BYTE(0)
    pcbNeeded = DWORD(0)
    pcReturned = DWORD(0)  # the number of PRINTER_INFO_1 structures retrieved
    r = EnumPrintersW(DWORD(flags), name, DWORD(level), byref(info), DWORD(0), byref(pcbNeeded), byref(pcReturned))
    log("EnumPrintersW(..)=%i pcbNeeded=%i", r, pcbNeeded.value)
    if pcbNeeded.value<=0:
        log("EnumPrinters probe failed for flags=%i, level=%i, pcbNeeded=%i", flags, level, pcbNeeded.value)
        return []

    bufsize = int(pcbNeeded.value)
    buf = (BYTE*bufsize)()
    pPrinterEnum = cast(addressof(buf), LPBYTE)
    log("EnumPrinters into buf %s", pPrinterEnum)
    r = EnumPrintersW(DWORD(flags), name, DWORD(level), pPrinterEnum, DWORD(bufsize), byref(pcbNeeded), byref(pcReturned))
    log("EnumPrintersW(..)=%i pcReturned=%i", r, pcReturned.value)
    if r==0:
        log.error("Error: EnumPrinters failed")
        return []
    info = cast(pPrinterEnum, POINTER(PRINTER_INFO))
    printers = []
    for i in range(pcReturned.value):
        v = int(info[i].Flags), str(info[i].pDescription), str(info[i].pName), str(info[i].pComment)
        log("EnumPrintersW(..) [%i]=%s", i, v)
        printers.append(v)
    del buf
    return printers

def get_info():
    from xpra.platform.printing import default_get_info
    i = default_get_info()
    #win32 extras:
    i.update({
        "skipped-printers"      : SKIPPED_PRINTERS,
        "printer-level"         : PRINTER_LEVEL,
        "default-printer-flags" : DEFAULT_PRINTER_FLAGS,
        "printer-flags"         : PRINTER_FLAGS,
        })
    return i

def get_printers():
    global PRINTER_ENUMS, PRINTER_ENUM_VALUES, SKIPPED_PRINTERS, PRINTER_LEVEL
    printers = {}
    for penum in PRINTER_ENUMS:
        try:
            eprinters = []
            enum_values = [PRINTER_ENUM_VALUES.get(x, 0) for x in penum]
            enum_val = sum(enum_values)
            log("enum(%s)=%s=%s", penum, "+".join(str(x) for x in enum_values), enum_val)
            assert enum_val is not None, "invalid printer enum %s" % penum
            log("querying %s printers with level=%s", penum, PRINTER_LEVEL)
            for p in EnumPrinters(enum_val, None, PRINTER_LEVEL):
                flags, desc, name, comment = p
                if name in SKIPPED_PRINTERS:
                    log("skipped printer: %#x, %s, %s, %s", flags, desc, name, comment)
                    continue
                if name in printers:
                    log("skipped duplicate printer: %#x, %s, %s, %s", flags, desc, name, comment)
                    continue
                log("found printer: %#x, %s, %s, %s", flags, desc, name, comment)
                #strip duplicated and empty strings from the description:
                desc_els = []
                for x in desc.split(","):
                    if x and not desc_els.count(x):
                        desc_els.append(x)
                info = {"printer-info"            : bytestostr(",".join(desc_els)),
                        "type"                    : penum}
                if comment:
                    info["printer-make-and-model"] = comment
                printers[name] = info
                eprinters.append(name)
            log("%s printers: %s", penum, eprinters)
        except Exception as e:
            log.warn("Warning: failed to query %s printers:", penum)
            log.warn(" %s", e)
            log("query error", exc_info=True)
            del e
    log("win32.get_printers()=%s", printers)
    return printers

def print_files(printer, filenames, title, options):
    log("win32.print_files%s", (printer, filenames, title, options))
    global JOB_ID, PROCESSES
    processes = []
    for filename in filenames:
        cwd = get_app_dir()
        command = ["PDFIUM_Print.exe", filename, printer, title]
        log("print command: %s", command)
        startupinfo = subprocess.STARTUPINFO()  # @UndefinedVariable
        startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW  # @UndefinedVariable
        startupinfo.wShowWindow = 0     #aka win32.con.SW_HIDE
        process = subprocess.Popen(command, stdout=sys.stdout, stderr=sys.stderr, cwd=cwd, startupinfo=startupinfo)
        process.print_filename = filename
        #we just let it run, no need for reaping the process on win32
        processes.append(process)
    JOB_ID +=1
    PROCESSES[JOB_ID] = processes
    log("win32.print_files(..)=%s (%s)", JOB_ID, processes)
    return JOB_ID

def printing_finished(jobid):
    global PROCESSES
    processes = PROCESSES.get(jobid)
    if not processes:
        log.warn("win32.printing_finished(%s) job not found!", jobid)
        return True
    log("win32.printing_finished(%s) processes: %s", jobid, [x.print_filename for x in processes])
    pending = [proc.print_filename for proc in processes if proc.poll() is None]
    log("win32.printing_finished(%s) still pending: %s", jobid, pending)
    #return finished when all the processes have terminated
    return len(pending)==0



# -*- coding: utf-8 -*-
# This file is part of Xpra.
# Copyright (C) 2020-2023 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.
#pylint: disable-msg=E1101

from collections import deque
from typing import Dict, Any, Deque

from xpra.server.mixins.stub_server_mixin import StubServerMixin
from xpra.net.common import PacketType


class ShellServer(StubServerMixin):
    """
    Mixin for adding `shell` support
    """
    def init(self, _opts) -> None:
        self.counter = 0
        self.commands : Deque[str] = deque(maxlen=10)

    def get_info(self, _source=None) -> Dict[str,Any]:
        return {
            "shell" : {
                "counter" : self.counter,
                "last-commands" : list(self.commands),
                },
            }

    def get_server_features(self, _source) -> Dict[str,Any]:
        return {
            "shell" : True,
            }

    def _process_shell_exec(self, proto, packet : PacketType) -> None:
        code = str(packet[1])
        ss = self.get_server_source(proto)
        if ss:
            self.counter += 1
            self.commands.append(code)
            ss.shell_exec(code)


    def init_packet_handlers(self) -> None:
        self.add_packet_handlers({
            "shell-exec" : self._process_shell_exec,
          }, False)



# This file is part of Xpra.
# Copyright (C) 2018-2023 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import re
import socket
from time import sleep, monotonic
from typing import Dict, List, Tuple, Any

from xpra.scripts.main import (
    InitException, InitExit,
    shellquote, host_target_string,
    )
from xpra.platform.paths import get_ssh_known_hosts_files
from xpra.platform.info import get_username
from xpra.scripts.config import parse_bool, TRUE_OPTIONS
from xpra.scripts.pinentry_wrapper import input_pass, confirm
from xpra.net.ssh.util import nogssapi_context, get_default_keyfiles
from xpra.net.bytestreams import SocketConnection, SOCKET_TIMEOUT
from xpra.make_thread import start_thread
from xpra.exit_codes import ExitCode
from xpra.os_util import (
    bytestostr, load_binary_file,
    umask_context,
    WIN32,
    )
from xpra.util import envint, envbool, envfloat, engs, noerr, csv, stderr_print
from xpra.log import Logger

#pylint: disable=import-outside-toplevel

log = Logger("network", "ssh")
if log.is_debug_enabled():
    import logging
    logging.getLogger("paramiko").setLevel(logging.DEBUG)


WINDOW_SIZE = envint("XPRA_SSH_WINDOW_SIZE", 2**27-1)
TIMEOUT = envint("XPRA_SSH_TIMEOUT", 60)

VERIFY_HOSTKEY = envbool("XPRA_SSH_VERIFY_HOSTKEY", True)
VERIFY_STRICT = envbool("XPRA_SSH_VERIFY_STRICT", False)
ADD_KEY = envbool("XPRA_SSH_ADD_KEY", True)
#which authentication mechanisms are enabled with paramiko:
NONE_AUTH = envbool("XPRA_SSH_NONE_AUTH", True)
PASSWORD_AUTH = envbool("XPRA_SSH_PASSWORD_AUTH", True)
AGENT_AUTH = envbool("XPRA_SSH_AGENT_AUTH", True)
KEY_AUTH = envbool("XPRA_SSH_KEY_AUTH", True)
PASSWORD_RETRY = envint("XPRA_SSH_PASSWORD_RETRY", 2)
SSH_AGENT = envbool("XPRA_SSH_AGENT", False)
assert PASSWORD_RETRY>=0
LOG_FAILED_CREDENTIALS = envbool("XPRA_LOG_FAILED_CREDENTIALS", False)
TEST_COMMAND_TIMEOUT = envint("XPRA_SSH_TEST_COMMAND_TIMEOUT", 10)
EXEC_STDOUT_TIMEOUT = envfloat("XPRA_SSH_EXEC_STDOUT_TIMEOUT", 2)
EXEC_STDERR_TIMEOUT = envfloat("XPRA_SSH_EXEC_STDERR_TIMEOUT", 0)

MSYS_DEFAULT_PATH = os.environ.get("XPRA_MSYS_DEFAULT_PATH", "/mingw64/bin/xpra")
CYGWIN_DEFAULT_PATH = os.environ.get("XPRA_CYGWIN_DEFAULT_PATH", "/cygdrive/c/Program Files/Xpra/Xpra_cmd.exe")
DEFAULT_WIN32_INSTALL_PATH = "C:\\Program Files\\Xpra"

PARAMIKO_SESSION_LOST = "No existing session"


def keymd5(k) -> str:
    import binascii
    f = bytestostr(binascii.hexlify(k.get_fingerprint()))
    s = "MD5"
    while f:
        s += ":"+f[:2]
        f = f[2:]
    return s


class SSHSocketConnection(SocketConnection):

    def __init__(self, ssh_channel, sock, sockname, peername, target, info=None, socket_options=None):
        self._raw_socket = sock
        super().__init__(ssh_channel, sockname, peername, target, "ssh", info, socket_options)

    def get_raw_socket(self):
        return self._raw_socket

    def start_stderr_reader(self) -> None:
        start_thread(self._stderr_reader, "ssh-stderr-reader", daemon=True)

    def _stderr_reader(self):
        #stderr = self._socket.makefile_stderr(mode="rb", bufsize=1)
        chan = self._socket
        stderr = chan.makefile_stderr("rb", 1)
        while self.active:
            v = stderr.readline()
            if not v:
                log.info("SSH EOF on stderr of %s", chan.get_name())
                break
            s = bytestostr(v.rstrip(b"\n\r"))
            if s:
                log.info(" SSH: %r", s)

    def peek(self, n) -> bytes:
        if not self._raw_socket:
            return b""
        return self._raw_socket.recv(n, socket.MSG_PEEK)

    def get_socket_info(self) -> Dict[str,Any]:
        if not self._raw_socket:
            return {}
        return self.do_get_socket_info(self._raw_socket)

    def get_info(self) -> Dict[str,Any]:
        i = super().get_info()
        s = self._socket
        if s:
            i["ssh-channel"] = {
                "id"    : s.get_id(),
                "name"  : s.get_name(),
                }
        return i


class SSHProxyCommandConnection(SSHSocketConnection):
    def __init__(self, ssh_channel, peername, target, info):
        super().__init__(ssh_channel, None, None, peername, target, info)
        self.process = None

    def error_is_closed(self, e) -> bool:
        p = self.process
        if p:
            #if the process has terminated,
            #then the connection must be closed:
            if p[0].poll() is not None:
                return True
        return super().error_is_closed(e)

    def get_socket_info(self) -> Dict[str,Any]:
        p = self.process
        if not p:
            return {}
        proc, _ssh, cmd = p
        return {
            "process" : {
                "pid"       : proc.pid,
                "returncode": proc.returncode,
                "command"   : cmd,
                }
            }

    def close(self) -> None:
        try:
            super().close()
        except Exception:
            #this can happen if the proxy command gets a SIGINT,
            #it's closed already and we don't care
            log("SSHProxyCommandConnection.close()", exc_info=True)


def safe_lookup(config_obj, host:str) -> Dict[Any,Any]:
    try:
        return config_obj.lookup(host) or {}
    except ImportError as e:
        log("%s.lookup(%s)", config_obj, host, exc_info=True)
        log.warn(f"Warning: unable to load SSH host config for {host!r}:")
        log.warn(f" {e}")
        if isinstance(e, ModuleNotFoundError):
            log.warn(" (looks like a 'paramiko' distribution packaging issue)")
    except KeyError as e:
        log("%s.lookup(%s)", config_obj, host, exc_info=True)
        log.info(f"paramiko ssh config lookup error for host {host!r}:")
        log.info(" %s: %s", type(e), e)
    return {}


def connect_to(display_desc):
    log(f"connect_to({display_desc})")
    #plain socket attributes:
    host = display_desc["host"]
    port = display_desc.get("port", 22)
    #ssh and command attributes:
    username = display_desc.get("username") or get_username()
    if "proxy_host" in display_desc:
        display_desc.setdefault("proxy_username", get_username())
    password = display_desc.get("password")
    remote_xpra = display_desc["remote_xpra"]
    proxy_command = display_desc["proxy_command"]       #ie: "_proxy_start"
    socket_dir = display_desc.get("socket_dir")
    display = display_desc.get("display")
    display_as_args = display_desc["display_as_args"]   #ie: "--start=xterm :10"
    paramiko_config = display_desc.copy()
    paramiko_config.update(display_desc.get("paramiko-config", {}))
    socket_info = {
            "host"  : host,
            "port"  : port,
            }
    def get_keyfiles(host_config, config_name="key"):
        keyfiles = (host_config or {}).get("identityfile") or get_default_keyfiles()
        keyfile = paramiko_config.get(config_name)
        if keyfile:
            keyfiles.insert(0, keyfile)
        return keyfiles

    def fail(msg:str) -> None:
        log("connect_to(%s)", display_desc, exc_info=True)
        raise InitExit(ExitCode.SSH_FAILURE, msg) from None

    with nogssapi_context():
        from paramiko import SSHConfig, ProxyCommand
        ssh_config = SSHConfig()
        def ssh_lookup(key):
            return safe_lookup(ssh_config, key)
        user_config_file = os.path.expanduser("~/.ssh/config")
        sock = None
        host_config = None
        if os.path.exists(user_config_file):
            with open(user_config_file, "r", encoding="utf8") as f:
                try:
                    ssh_config.parse(f)
                except Exception as e:
                    log(f"parse({user_config_file})", exc_info=True)
                    log.error(f"Error parsing {user_config_file!r}:")
                    log.estr(e)
            log(f"parsed user config {user_config_file!r}")
            try:
                log("%i hosts found", len(ssh_config.get_hostnames()))
            except KeyError:
                pass
            host_config = ssh_lookup(host)
            if host_config:
                log(f"got host config for {host!r}: {host_config}")
                host = host_config.get("hostname", host)
                username = host_config.get("user", username)
                port = host_config.get("port", port)
                try:
                    port = int(port)
                except (TypeError, ValueError):
                    raise InitExit(ExitCode.SSH_FAILURE, f"invalid ssh port specified: {port!r}") from None
                proxycommand = host_config.get("proxycommand")
                if proxycommand:
                    log(f"found proxycommand={proxycommand!r} for host {host!r}")
                    sock = ProxyCommand(proxycommand)
                    log(f"ProxyCommand({proxycommand})={sock}")
                    from xpra.child_reaper import getChildReaper
                    cmd = getattr(sock, "cmd", [])
                    def proxycommand_ended(proc):
                        log(f"proxycommand_ended({proc}) exit code={proc.poll()}")
                    getChildReaper().add_process(sock.process, "paramiko-ssh-client", cmd, True, True,
                                                 callback=proxycommand_ended)
                    proxy_keys = get_keyfiles(host_config, "proxy_key")
                    log(f"proxy keys={proxy_keys}")
                    from paramiko.client import SSHClient
                    ssh_client = SSHClient()
                    ssh_client.load_system_host_keys()
                    log("ssh proxy command connect to %s", (host, port, sock))
                    ssh_client.connect(host, port, sock=sock)
                    transport = ssh_client.get_transport()
                    do_connect_to(transport, host,
                                               username, password,
                                               host_config or ssh_lookup("*"),
                                               proxy_keys,
                                               paramiko_config)
                    chan = run_remote_xpra(transport, proxy_command, remote_xpra, socket_dir, display_as_args)
                    peername = (host, port)
                    conn = SSHProxyCommandConnection(chan, peername, peername, socket_info)
                    conn.target = host_target_string("ssh", username, host, port, display)
                    conn.timeout = SOCKET_TIMEOUT
                    conn.start_stderr_reader()
                    conn.process = (sock.process, "ssh", cmd)
                    from xpra.net import bytestreams
                    from paramiko.ssh_exception import ProxyCommandFailure
                    bytestreams.CLOSED_EXCEPTIONS = tuple(list(bytestreams.CLOSED_EXCEPTIONS)+[ProxyCommandFailure])
                    return conn

        keys = get_keyfiles(host_config)
        from xpra.net.socket_util import socket_connect
        if "proxy_host" in display_desc:
            proxy_host = display_desc["proxy_host"]
            proxy_port = int(display_desc.get("proxy_port", 22))
            proxy_username = display_desc.get("proxy_username", username)
            proxy_password = display_desc.get("proxy_password", password)
            proxy_keys = get_keyfiles(host_config, "proxy_key")
            sock = socket_connect(proxy_host, proxy_port)
            if not sock:
                fail(f"SSH proxy transport failed to connect to {proxy_host}:{proxy_port}")
            middle_transport = do_connect(sock, proxy_host,
                                                       proxy_username, proxy_password,
                                                       ssh_lookup(host) or ssh_lookup("*"),
                                                       proxy_keys,
                                                       paramiko_config)
            log("Opening proxy channel")
            chan_to_middle = middle_transport.open_channel("direct-tcpip", (host, port), ("localhost", 0))
            transport = do_connect(chan_to_middle, host,
                                                username, password,
                                                host_config or ssh_lookup("*"),
                                                keys,
                                                paramiko_config)
            chan = run_remote_xpra(transport, proxy_command, remote_xpra, socket_dir, display_as_args)
            peername = (host, port)
            conn = SSHProxyCommandConnection(chan, peername, peername, socket_info)
            conn.target = host_target_string("ssh", username, host, port, display) \
                            + " via " + \
                            host_target_string("ssh", proxy_username, proxy_host, proxy_port, None)
            conn.timeout = SOCKET_TIMEOUT
            conn.start_stderr_reader()
            return conn

        #plain TCP connection to the server,
        #we open it then give the socket to paramiko:
        auth_modes = get_auth_modes(paramiko_config, host_config, password)
        log(f"authentication modes={auth_modes}")
        sock = None
        transport = None
        while not transport:
            sock = socket_connect(host, port)
            if not sock:
                fail(f"SSH failed to connect to {host}:{port}")
            sockname = sock.getsockname()
            peername = sock.getpeername()
            log(f"paramiko socket_connect: sockname={sockname}, peername={peername}")
            try:
                transport = do_connect(sock, host, username, password,
                                                    host_config or ssh_lookup("*"),
                                                    keys,
                                                    paramiko_config,
                                                    auth_modes)
            except SSHAuthenticationError as e:
                log(f"paramiko authentication errors on socket {sock} with modes {auth_modes}: {e.errors}",
                    exc_info=True)
                pw_errors = []
                for errs in e.errors.values():
                    pw_errors += errs
                if ("key" in auth_modes or "agent" in auth_modes) and PARAMIKO_SESSION_LOST in pw_errors:
                    #try connecting again but without 'key' and 'agent' authentication:
                    #see https://github.com/Xpra-org/xpra/issues/3223
                    for m in ("key", "agent"):
                        try:
                            auth_modes.remove(m)
                        except KeyError:
                            pass
                    log.info(f"retrying SSH authentication with modes {csv(auth_modes)}")
                    continue
                raise
            finally:
                if sock and not transport:
                    noerr(sock.shutdown)
                    noerr(sock.close)
                    sock = None

        remote_port = display_desc.get("remote_port", 0)
        if remote_port:
            #we want to connect directly to a remote port,
            #we don't need to run a command
            chan = transport.open_channel("direct-tcpip", ("localhost", remote_port), ('localhost', 0))
            log(f"direct channel to remote port {remote_port} : {chan}")
        else:
            chan = run_remote_xpra(transport, proxy_command, remote_xpra,
                                            socket_dir, display_as_args, paramiko_config)
        conn = SSHSocketConnection(chan, sock, sockname, peername, (host, port), socket_info)
        conn.target = host_target_string("ssh", username, host, port, display)
        conn.timeout = SOCKET_TIMEOUT
        conn.start_stderr_reader()
        return conn


AUTH_MODES : Tuple[str, ...] = ("none", "agent", "key", "password")

def get_auth_modes(paramiko_config, host_config, password) -> List[str]:
    def configvalue(key):
        #if the paramiko config has a setting, honour it:
        if paramiko_config and key in paramiko_config:
            return paramiko_config.get(key)
        #fallback to the value from the host config:
        return (host_config or {}).get(key)
    def configbool(key, default_value=True):
        return parse_bool(key, configvalue(key), default_value)
    auth_str = configvalue("auth")
    if auth_str:
        return auth_str.split("+")
    auth = []
    if configbool("noneauthentication", NONE_AUTH):
        auth.append("none")
    if password and configbool("passwordauthentication", PASSWORD_AUTH):
        auth.append("password")
    if configbool("agentauthentication", AGENT_AUTH):
        auth.append("agent")
    # Some people do two-factor using KEY_AUTH to kick things off, so this happens first
    if configbool("keyauthentication", KEY_AUTH):
        auth.append("key")
    if not password and configbool("passwordauthentication", PASSWORD_AUTH):
        auth.append("password")
    return auth


class iauthhandler:
    def __init__(self, password):
        self.authcount = 0
        self.password = password
    def handle_request(self, title:str, instructions, prompt_list) -> List:
        log("handle_request%s counter=%i", (title, instructions, prompt_list), self.authcount)
        p = []
        for pent in prompt_list:
            if self.password:
                p.append(self.password)
                self.password = None
            else:
                p.append(input_pass(pent[0]))
        self.authcount += 1
        log(f"handle_request(..) returning {len(p)} values")
        return p


def do_connect(chan, host:str, username:str, password:str,
                            host_config=None, keyfiles=None, paramiko_config=None, auth_modes=AUTH_MODES):
    from paramiko import SSHException
    from paramiko.transport import Transport
    transport = Transport(chan)
    transport.use_compression(False)
    log("SSH transport %s", transport)
    try:
        transport.start_client()
    except SSHException as e:
        log("SSH negotiation failed", exc_info=True)
        raise InitExit(ExitCode.SSH_FAILURE, "SSH negotiation failed: %s" % e) from None
    return do_connect_to(transport, host, username, password,
                                      host_config, keyfiles, paramiko_config, auth_modes)

def do_connect_to(transport, host:str, username:str, password:str,
                               host_config=None, keyfiles=None, paramiko_config=None, auth_modes=AUTH_MODES):
    from paramiko import SSHException, PasswordRequiredException
    from paramiko.agent import Agent
    from paramiko.hostkeys import HostKeys
    log("do_connect_to%s", (transport, host, username, password, host_config, keyfiles, paramiko_config))

    def configvalue(key):
        #if the paramiko config has a setting, honour it:
        if paramiko_config and key in paramiko_config:
            return paramiko_config.get(key)
        #fallback to the value from the host config:
        return (host_config or {}).get(key)
    def configbool(key, default_value=True) -> bool:
        return parse_bool(key, configvalue(key), default_value)
    def configint(key, default_value=0) -> int:
        v = configvalue(key)
        if v is None:
            return default_value
        return int(v)

    host_key = transport.get_remote_server_key()
    assert host_key, "no remote server key"
    log("remote_server_key=%s", keymd5(host_key))
    if configbool("verify-hostkey", VERIFY_HOSTKEY):
        host_keys = HostKeys()
        host_keys_filename = None
        KNOWN_HOSTS = get_ssh_known_hosts_files()
        for known_hosts in KNOWN_HOSTS:
            host_keys.clear()
            try:
                path = os.path.expanduser(known_hosts)
                if os.path.exists(path):
                    host_keys.load(path)
                    log("HostKeys.load(%s) successful", path)
                    host_keys_filename = path
                    break
            except IOError:
                log("HostKeys.load(%s)", known_hosts, exc_info=True)

        log("host keys=%s", host_keys)
        keys = safe_lookup(host_keys, host)
        known_host_key = keys.get(host_key.get_name())
        def keyname():
            return host_key.get_name().replace("ssh-", "")
        if known_host_key and host_key==known_host_key:
            assert host_key
            log("%s host key '%s' OK for host '%s'", keyname(), keymd5(host_key), host)
        else:
            dnscheck = ""
            if configbool("verifyhostkeydns"):
                try:
                    from xpra.net.ssh.sshfp import check_host_key
                    dnscheck = check_host_key(host, host_key)
                except ImportError as e:
                    log("verifyhostkeydns failed", exc_info=True)
                    log.info("cannot check SSHFP DNS records")
                    log.info(" %s", e)
            log("dnscheck=%s", dnscheck)
            def adddnscheckinfo(q : List[str]):
                if dnscheck is not True:
                    if dnscheck:
                        q.append("SSHFP validation failed:")
                        q.append(str(dnscheck))
                    else:
                        q.append("SSHFP validation failed")
            if dnscheck is True:
                #DNSSEC provided a matching record
                log.info("found a valid SSHFP record for host %s", host)
            elif known_host_key:
                log.warn("Warning: SSH server key mismatch")
                qinfo : List[str] = [
                    "WARNING: REMOTE HOST IDENTIFICATION HAS CHANGED!",
                    "IT IS POSSIBLE THAT SOMEONE IS DOING SOMETHING NASTY!",
                    "Someone could be eavesdropping on you right now (man-in-the-middle attack)!",
                    "It is also possible that a host key has just been changed.",
                    f"The fingerprint for the {keyname()} key sent by the remote host is",
                    keymd5(host_key),
                    ]
                adddnscheckinfo(qinfo)
                if configbool("stricthostkeychecking", VERIFY_STRICT):
                    log.warn("Host key verification failed.")
                    #TODO: show alert with no option to accept key
                    qinfo += [
                        "Please contact your system administrator.",
                        "Add correct host key in %s to get rid of this message.",
                        f"Offending {keyname()} key in {host_keys_filename}",
                        f"ECDSA host key for {keyname()} has changed and you have requested strict checking.",
                        ]
                    stderr_print(os.linesep.join(qinfo))
                    transport.close()
                    raise InitExit(ExitCode.SSH_KEY_FAILURE, "SSH Host key has changed")
                if not confirm(qinfo):
                    transport.close()
                    raise InitExit(ExitCode.SSH_KEY_FAILURE, "SSH Host key has changed")
                log.info("host key confirmed")
            else:
                assert (not keys) or (host_key.get_name() not in keys)
                if not keys:
                    log.warn("Warning: unknown SSH host '%s'", host)
                else:
                    log.warn("Warning: unknown %s SSH host key", keyname())
                qinfo = [
                    f"The authenticity of host {host!r} can't be established.",
                    f"{keyname()} key fingerprint is",
                    keymd5(host_key),
                    ]
                adddnscheckinfo(qinfo)
                if not confirm(qinfo):
                    transport.close()
                    raise InitExit(ExitCode.SSH_KEY_FAILURE, f"Unknown SSH host {host!r}")
                log.info("host key confirmed")
            if configbool("addkey", ADD_KEY):
                try:
                    if not host_keys_filename:
                        #the first one is the default,
                        #ie: ~/.ssh/known_hosts on posix
                        host_keys_filename = os.path.expanduser(KNOWN_HOSTS[0])
                    log(f"adding {keyname()} key for host {host!r} to {host_keys_filename!r}")
                    if not os.path.exists(host_keys_filename):
                        keys_dir = os.path.dirname(host_keys_filename)
                        if not os.path.exists(keys_dir):
                            log(f"creating keys directory {keys_dir!r}")
                            os.mkdir(keys_dir, 0o700)
                        elif not os.path.isdir(keys_dir):
                            log.warn(f"Warning: {keys_dir!r} is not a directory")
                            log.warn(" key not saved")
                        if os.path.exists(keys_dir) and os.path.isdir(keys_dir):
                            log(f"creating known host file {host_keys_filename!r}")
                            with umask_context(0o133):
                                with open(host_keys_filename, "ab+"):
                                    "file has been created"
                    host_keys.add(host, host_key.get_name(), host_key)
                    host_keys.save(host_keys_filename)
                except OSError as e:
                    log(f"failed to add key to {host_keys_filename!r}")
                    log.error(f"Error adding key to {host_keys_filename!r}")
                    log.error(f" {e}")
                except Exception:
                    log.error("cannot add key", exc_info=True)
    else:
        log("ssh host key verification skipped")


    auth_errors : Dict[str,List[str]] = {}

    def auth_agent() -> None:
        agent = Agent()
        agent_keys = agent.get_keys()
        log("agent keys: %s", agent_keys)
        if agent_keys:
            for agent_key in agent_keys:
                log("trying ssh-agent key '%s'", keymd5(agent_key))
                try:
                    transport.auth_publickey(username, agent_key)
                    if transport.is_authenticated():
                        log("authenticated using agent and key '%s'", keymd5(agent_key))
                        return
                except AttributeError as e:
                    log(f"auth_publickey({username}, {agent_key})")
                    log.warn("Warning: paramiko bug during public key agent authentication")
                    log.warn(f" {type(e)}: {e}")
                    log.warn(f" using key {type(agent_key)}: {agent_key}")
                except SSHException as e:
                    auth_errors.setdefault("agent", []).append(str(e))
                    log.info("SSH agent key '%s' rejected for user '%s'", keymd5(agent_key), username)
                    log("%s%s", transport.auth_publickey, (username, agent_key), exc_info=True)
                    if str(e)==PARAMIKO_SESSION_LOST:
                        #no point in trying more keys
                        break
            if not transport.is_authenticated():
                log.info("agent authentication failed, tried %i key%s", len(agent_keys), engs(agent_keys))

    def auth_publickey() -> None:
        log(f"trying public key authentication using {keyfiles}")
        for keyfile_path in keyfiles:
            if not os.path.exists(keyfile_path):
                log(f"no keyfile at {keyfile_path!r}")
                continue
            log(f"trying {keyfile_path!r}")
            key = None
            import paramiko
            try_key_formats = ()
            for kf in ("RSA", "DSS", "ECDSA", "Ed25519"):
                if keyfile_path.lower().endswith(kf.lower()):
                    try_key_formats = (kf, )
                    break
            if not try_key_formats:
                try_key_formats = ("RSA", "DSS", "ECDSA", "Ed25519")
            pkey_classname = None
            for pkey_classname in try_key_formats:
                pkey_class = getattr(paramiko, f"{pkey_classname}Key", None)
                if pkey_class is None:
                    log(f"no {pkey_classname} key type")
                    continue
                log(f"trying to load as {pkey_classname}")
                key = None
                try:
                    key = pkey_class.from_private_key_file(keyfile_path)
                    log.info(f"loaded {pkey_classname} private key from {keyfile_path!r}")
                    break
                except PasswordRequiredException as e:
                    log(f"{keyfile_path!r} keyfile requires a passphrase: {e}")
                    passphrase = input_pass(f"please enter the passphrase for {keyfile_path!r}")
                    if passphrase:
                        try:
                            key = pkey_class.from_private_key_file(keyfile_path, passphrase)
                            log.info(f"loaded {pkey_classname} private key from {keyfile_path!r}")
                        except SSHException as ke:
                            log("from_private_key_file", exc_info=True)
                            log.info(f"cannot load key from file {keyfile_path}:")
                            for emsg in str(ke).split(". "):
                                log.info(" %s.", emsg)
                    break
                except Exception:
                    log(f"auth_publickey() loading as {pkey_classname}", exc_info=True)
                    key_data = load_binary_file(keyfile_path)
                    if key_data and key_data.find(b"BEGIN OPENSSH PRIVATE KEY")>=0 and paramiko.__version__<"2.7":
                        log.warn(f"Warning: private key {keyfile_path!r}")
                        log.warn(" this file seems to be using OpenSSH's own format")
                        log.warn(" please convert it to something more standard (ie: PEM)")
                        log.warn(" so it can be used with the paramiko backend")
                        if WIN32:
                            log.warn(" or switch to the Putty Plink backend with:")
                            log.warn(" '--ssh=plink -ssh -agent'")
                        else:
                            log.warn(" or switch to the OpenSSH backend with:")
                            log.warn(" '--ssh=ssh'")
            if key:
                log(f"auth_publickey using {keyfile_path!r} as {pkey_classname}: {keymd5(key)}")
                try:
                    transport.auth_publickey(username, key)
                except SSHException as e:
                    auth_errors.setdefault("key", []).append(str(e))
                    log(f"key {keyfile_path!r} rejected", exc_info=True)
                    log.info(f"SSH authentication using key {keyfile_path!r} failed:")
                    log.info(f" {e}")
                    if str(e)==PARAMIKO_SESSION_LOST:
                        #no point in trying more keys
                        break
                else:
                    if transport.is_authenticated():
                        break
            else:
                log.error(f"Error: cannot load private key {keyfile_path!r}")

    def auth_none() -> None:
        log("trying none authentication")
        try:
            transport.auth_none(username)
        except SSHException as e:
            auth_errors.setdefault("none", []).append(str(e))
            log("auth_none()", exc_info=True)

    def auth_password() -> None:
        log("trying password authentication")
        try:
            transport.auth_password(username, password)
        except SSHException as e:
            auth_errors.setdefault("password", []).append(str(e))
            log("auth_password(..)", exc_info=True)
            emsgs = getattr(e, "message", str(e)).split(";")
        else:
            emsgs = []
        if not transport.is_authenticated():
            log.info("SSH password authentication failed:")
            for emsg in emsgs:
                log.info(f" {emsg}")
            if log.is_debug_enabled() and LOG_FAILED_CREDENTIALS:
                log.info(f" invalid username {username!r} or password {password!r}")

    def auth_interactive() -> None:
        log("trying interactive authentication")
        try:
            myiauthhandler = iauthhandler(password)
            transport.auth_interactive(username, myiauthhandler.handle_request, "")
        except SSHException as e:
            auth_errors.setdefault("interactive", []).append(str(e))
            log("auth_interactive(..)", exc_info=True)
            log.info("SSH password authentication failed:")
            for emsg in getattr(e, "message", str(e)).split(";"):
                log.info(f" {emsg}")
        finally:
            del myiauthhandler

    banner = transport.get_banner()
    if banner:
        log.info("SSH server banner:")
        for x in banner.splitlines():
            log.info(f" {x}")

    log(f"starting authentication, authentication methods: {auth_modes}")
    auth = list(auth_modes)
    # per the RFC we probably should do none first always and read off the supported
    # methods, however, the current code seems to work fine with OpenSSH
    while not transport.is_authenticated() and auth:
        a = auth.pop(0)
        log("auth=%s", a)
        if a=="none":
            auth_none()
        elif a=="agent":
            auth_agent()
        elif a=="key":
            auth_publickey()
        elif a=="password":
            auth_interactive()
            if not transport.is_authenticated():
                if password:
                    auth_password()
                else:
                    tries = configint("numberofpasswordprompts", PASSWORD_RETRY)
                    for _ in range(tries):
                        password = input_pass(f"please enter the SSH password for {username}@{host}")
                        if not password:
                            break
                        auth_password()
                        if transport.is_authenticated():
                            break
        else:
            log.warn(f"Warning: invalid authentication mechanism {a}")
        #detect session-lost problems:
        #(no point in continuing without a session)
        if auth_errors and not transport.is_authenticated():
            for err_strs in auth_errors.values():
                if PARAMIKO_SESSION_LOST in err_strs:
                    raise SSHAuthenticationError(host, auth_errors)
    if not transport.is_authenticated():
        transport.close()
        log(f"authentication errors: {auth_errors}")
        raise SSHAuthenticationError(host, auth_errors)
    return transport

class SSHAuthenticationError(InitExit):
    def __init__(self, host, errors):
        super().__init__(ExitCode.CONNECTION_FAILED, f"SSH Authentication failed for {host!r}")
        self.errors = errors

def run_test_command(transport, cmd:str) -> Tuple[bytes,bytes,int]:
    from paramiko import SSHException
    log(f"run_test_command(transport, {cmd})")
    try:
        chan = transport.open_session(window_size=None, max_packet_size=0, timeout=60)
        chan.set_name(f"run-test:{cmd}")
    except SSHException as e:
        log("open_session", exc_info=True)
        raise InitExit(ExitCode.SSH_FAILURE, f"failed to open SSH session: {e}") from None
    chan.exec_command(cmd)
    log(f"exec_command({cmd!r}) returned")
    start = monotonic()
    while not chan.exit_status_ready():
        if monotonic()-start>TEST_COMMAND_TIMEOUT:
            chan.close()
            raise InitException(f"SSH test command {cmd!r} timed out")
        log("exit status is not ready yet, sleeping")
        sleep(0.01)
    code = chan.recv_exit_status()
    log(f"exec_command({cmd!r})={code}")
    def chan_read(fileobj):
        try:
            return fileobj.readlines()
        except OSError:
            log(f"chan_read({fileobj})", exc_info=True)
            return []
        finally:
            noerr(fileobj.close)
    #don't wait too long for the data:
    chan.settimeout(EXEC_STDOUT_TIMEOUT)
    out = chan_read(chan.makefile())
    log(f"exec_command out={out!r}")
    chan.settimeout(EXEC_STDERR_TIMEOUT)
    err = chan_read(chan.makefile_stderr())
    log(f"exec_command err={err!r}")
    chan.close()
    return out, err, code


def run_remote_xpra(transport, xpra_proxy_command=None, remote_xpra=None,
                             socket_dir=None, display_as_args=None, paramiko_config=None):
    log("run_remote_xpra%s", (transport, xpra_proxy_command, remote_xpra,
                             socket_dir, display_as_args, paramiko_config))
    from paramiko import SSHException
    assert remote_xpra
    log(f"will try to run xpra from: {remote_xpra}")
    def rtc(cmd):
        return run_test_command(transport, cmd)
    def detectosname() -> str:
        #first, try a syntax that should work with any ssh server:
        r = rtc("echo %OS%")
        if r[2]==0 and r[0]:
            name = r[0][-1].rstrip("\n\r")
            log(f"echo %OS%={name!r}")
            if name!="%OS%":
                #MS Windows OS will return "Windows_NT" here
                log.info(f"ssh server OS is {name!r}")
                return name
        #this should work on all other OSes:
        r = rtc("echo $OSTYPE")
        if r[2]==0 and r[0]:
            name = r[0][-1].rstrip("\n\r")
            log(f"OSTYPE={name!r}")
            log.info(f"ssh server OS is {name!r}")
            return name
        return "unknown"
    WIN32_REGISTRY_QUERY = "REG QUERY \"HKEY_LOCAL_MACHINE\\Software\\Xpra\" /v InstallPath"
    def getexeinstallpath():
        cmd = WIN32_REGISTRY_QUERY
        if osname=="msys":
            #escape for msys shell:
            cmd = cmd.replace("/", "//")
        r = rtc(cmd)
        if r[2]!=0:
            return None
        out = r[0]
        for line in out:
            qmatch = re.search(r"InstallPath\s*\w*\s*(.*)", line)
            if qmatch:
                return qmatch.group(1).rstrip("\n\r")
        return None
    osname = detectosname()
    tried = set()
    find_command = None
    for xpra_cmd in remote_xpra:
        found = False
        if osname.startswith("Windows") or osname in ("msys", "cygwin"):
            #on MS Windows,
            #always prefer the application path found in the registry:
            def winpath(p) -> str:
                if osname=="msys":
                    return p.replace("\\", "\\\\")
                if osname=="cygwin":
                    return "/cygdrive/"+p.replace(":\\", "/").replace("\\", "/")
                return p
            installpath = getexeinstallpath()
            if installpath:
                xpra_cmd = winpath(f"{installpath}\\Xpra_cmd.exe")
                found = True
            elif xpra_cmd.find("/")<0 and xpra_cmd.find("\\")<0:
                test_path = winpath(f"{DEFAULT_WIN32_INSTALL_PATH}\\{xpra_cmd}")
                cmd = f'dir "{test_path}"'
                r = rtc(cmd)
                if r[2]==0:
                    xpra_cmd = test_path
                    found = True
        if not found and not find_command and not osname.startswith("Windows"):
            if rtc("command")[2]==0:
                find_command = "command -v"
            else:
                find_command = "which"
        if not found and find_command:
            r = rtc(f"{find_command} {xpra_cmd}")
            out = r[0]
            if r[2]==0 and out:
                #use the actual path returned by 'command -v' or 'which':
                try:
                    xpra_cmd = out[-1].rstrip("\n\r ").lstrip("\t ")
                except Exception as e:
                    log(f"cannot get command from {xpra_cmd}: {e}")
                else:
                    if xpra_cmd.startswith(f"alias {xpra_cmd}="):
                        #ie: "alias xpra='xpra -d proxy'" -> "xpra -d proxy"
                        xpra_cmd = xpra_cmd.split("=", 1)[1].strip("'")
                    found = bool(xpra_cmd)
            elif xpra_cmd=="xpra" and osname in ("msys", "cygwin"):
                default_path = CYGWIN_DEFAULT_PATH if osname=="cygwin" else MSYS_DEFAULT_PATH
                if default_path:
                    #try the default system installation path
                    r = rtc(f"command -v '{default_path}'")
                    if r[2]==0:
                        xpra_cmd = default_path     #ie: "/mingw64/bin/xpra"
                        found = True
        if not found or xpra_cmd in tried:
            continue
        log(f"adding xpra_cmd={xpra_cmd!r}")
        tried.add(xpra_cmd)
        cmd = '"' + xpra_cmd + '" ' + ' '.join(shellquote(x) for x in xpra_proxy_command)
        if socket_dir:
            cmd += f" \"--socket-dir={socket_dir}\""
        if display_as_args:
            cmd += " "
            cmd += " ".join(shellquote(x) for x in display_as_args)
        log(f"cmd({xpra_proxy_command}, {display_as_args})={cmd}")

        #see https://github.com/paramiko/paramiko/issues/175
        #WINDOW_SIZE = 2097152
        log(f"trying to open SSH session, window-size={WINDOW_SIZE}, timeout={TIMEOUT}")
        try:
            chan = transport.open_session(window_size=WINDOW_SIZE, max_packet_size=0, timeout=TIMEOUT)
            chan.set_name("run-xpra")
        except SSHException as e:
            log("open_session", exc_info=True)
            raise InitExit(ExitCode.SSH_FAILURE, f"failed to open SSH session: {e}") from None
        else:
            agent_option = str((paramiko_config or {}).get("agent", SSH_AGENT)) or "no"
            log(f"paramiko agent_option={agent_option}")
            if agent_option.lower() in TRUE_OPTIONS:
                log.info("paramiko SSH agent forwarding enabled")
                from paramiko.agent import AgentRequestHandler
                AgentRequestHandler(chan)
            log(f"channel exec_command({cmd!r})")
            chan.exec_command(cmd)
            return chan
    raise RuntimeError("all SSH remote proxy commands have failed - is xpra installed on the remote host?")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# This file is part of Xpra.
# Copyright (C) 2020-2022 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import unittest

from xpra.os_util import nomodule_context, WIN32
from xpra.scripts.parsing import (
    parse_ssh_option, get_ssh_args, get_ssh_proxy_args, parse_remote_display,
    )

class TestParsing(unittest.TestCase):

    def test_ssh_parsing(self):
        assert parse_ssh_option("auto")[0] in ("paramiko", "ssh")
        assert parse_ssh_option("ssh")==["ssh"]
        assert parse_ssh_option("ssh -v")==["ssh", "-v"]
        with nomodule_context("paramiko"):
            def pssh(s, e):
                r = parse_ssh_option(s)[0]
                assert r==e, f"expected {e} got {r}"
            if WIN32:
                pssh("auto", "plink.exe")
            else:
                pssh("auto", "ssh")
        #args:
        def targs(e, *args):
            r = get_ssh_args(*args)
            assert r==e, f"expected {e} but got {r}"
        targs([], {"host" : "host"})
        targs(["-pw", "password1", "-l", "username1", "-P", "2222", "-T", "host1"], {
            "username" : "username1",
            "password" : "password1",
            "host"    : "host1",
            "port"    : 2222,
            }, ["putty"])
        if not WIN32:
            keyfile = os.path.expanduser("~/key")
            targs(["-l", "username1", "-p", "2222", "-T", "host1", "-i", keyfile], {
                "username" : "username1",
                "password" : "password1",
                "host"    : "host1",
                "port"    : 2222,
                "key"     : keyfile,
                }, ["ssh"])
        #ssh proxy:
        def pargs(e, n, *args):
            r = get_ssh_proxy_args(*args)[:n]
            assert r==e, f"expected {e} but got {r}"
        pargs(["-o"], 1, {
            "proxy_username" : "username1",
            "proxy_password" : "password1",
            "proxy_host"    : "host1",
            "proxy_port"    : 2222,
            }, ["ssh"])
        pargs(["-proxycmd"], 1, {
            "proxy_username" : "username1",
            "proxy_password" : "password1",
            "proxy_host"    : "host1",
            "proxy_port"    : 2222,
            }, ["plink"])
        #remote display attributes:
        assert parse_remote_display("somedisplay").get("display")=="somedisplay"
        assert parse_remote_display("10?proxy=username:password@host:222").get("proxy")=="username:password@host:222"
        def t(s):
            v = parse_remote_display(s)
            assert v.get("display")=="somedisplay"
        t("somedisplay?proxy=")
        t("somedisplay?proxy=:22")
        t("somedisplay?proxy=:@host:22")
        t("somedisplay?proxy=:password@host:22")


def main():
    unittest.main()

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2013-2023 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Tuple

from xpra.platform.win32.auth import check
from xpra.server.auth.sys_auth_base import SysAuthenticator, log


class Authenticator(SysAuthenticator):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        #fugly: keep hold of the password so the win32 proxy can use it
        self.password = ""

    def get_uid(self) -> int:
        return 0

    def get_gid(self) -> int:
        return 0

    def get_password(self) -> str:
        return self.password

    def get_challenge(self, digests) -> Tuple[bytes,str]:
        self.req_xor(digests)
        return super().do_get_challenge(["xor"])

    def check(self, password:bytes) -> bool:
        domain = '' #os.environ.get('COMPUTERNAME')
        if check(domain, self.username, password):
            self.password = password
            return True
        return False

    def __repr__(self):
        return "win32"


def main(argv) -> int:
    #pylint: disable=import-outside-toplevel
    from xpra.platform import program_context
    from xpra.log import enable_color
    with program_context("Auth-Test", "Auth-Test"):
        enable_color()
        for x in ("-v", "--verbose"):
            if x in tuple(argv):
                log.enable_debug()
                argv.remove(x)
        if len(argv)!=3:
            log.warn("invalid number of arguments")
            log.warn("usage: %s [--verbose] username password", argv[0])
            return 1
        username = argv[1]
        password = argv[2]
        if check("", username, password):
            log.info("authentication succeeded")
            return 0
        log.error("authentication failed")
        return 1

if __name__ == "__main__":
    import sys
    sys.exit(main(sys.argv))
